"""Functions for splitting channels and or slices."""

import os

from ij import IJ, ImagePlus  # pylint: disable-msg=E0401
from ij.io import FileSaver  # pylint: disable-msg=E0401
from ij.plugin import ChannelSplitter  # pylint: disable-msg=E0401


def split_by_c_and_z(log, dname, imgf, skip_top, skip_bottom):
    """Open a file, split by Z and C and save the result into individual TIFFs.

    Load the file specified, split by channels and z-slices, create a directory
    for each channel using the channel number as a name suffix and export
    each slice as an individual TIF file.

    Parameters
    ----------
    log : logger or scijava-logservice
        The logger object to be used for logging.
    dname : str
        The directory to load TIF files from.
    imgf : str
        The file name to load and split.
    skip_top : int
        Number of slices to skip at the top.
    skip_bottom : int
        Number of slices to skip at the bottom.
    """
    log.info("Processing file [%s]" % imgf)
    imp = IJ.openImage(dname + "/" + imgf)
    fname = os.path.splitext(imgf)
    channels = ChannelSplitter().split(imp)
    for channel in channels:
        c_name = channel.getTitle().split("-")[0]
        tgt_dir = os.path.join(dname, fname[0] + "-" + c_name)
        if not os.path.isdir(tgt_dir):
            os.mkdir(tgt_dir)
        stack = channel.getStack()
        for z in range(1 + skip_top, stack.getSize() + 1 - skip_bottom):
            proc = stack.getProcessor(z)
            fout = "%s/%s-z%s%s" % (tgt_dir, fname[0], z, fname[1])
            # fout = dname + "/" + c_name + "/" + fname[0] + "-z" + z + fname[1]
            log.info("Writing channel %s, slice %s: %s" % (c_name, z, fout))
            FileSaver(ImagePlus(fname[0], proc)).saveAsTiff(fout)



"""I/O related functions."""

import zipfile

from os.path import splitext, join

from .log import LOG as log
from .strtools import flatten

from ._jython_compat import file_types


def filehandle(fname, mode="r"):
    """Make sure a variable is either a filehandle or create one from it.

    This function takes a variable and checks whether it is already a
    filehandle with the desired mode or a string that can be turned into a
    filehandle with that mode. This can be used e.g. to make functions agnostic
    against being supplied a file-type parameter that was gathered via argparse
    (then it's already a filehandle) or as a plain string.

    Parameters
    ----------
    fname : str or filehandle
        The object to ensure it is a file handle, or to create one from it.
    mode : str
        The desired mode of the filehandle (default=read).

    Returns
    -------
    A valid (open) filehandle with the given mode. Raises an IOError
    otherwise.

    Example
    -------
    >>> fname = __file__
    >>> type(fname)
    <type 'str'>
    >>> type(filehandle(fname))
    <type 'file'>
    >>> fh = open(__file__, 'r')
    >>> type(fh)
    <type 'file'>
    >>> type(filehandle(fh))
    <type 'file'>
    """
    log.debug(type(fname))
    if isinstance(fname, str):
        try:
            return open(fname, mode)
        except IOError as err:
            message = "can't open '%s': %s"
            raise SystemExit(message % (fname, err))
    elif isinstance(fname, file_types):
        if fname.mode != mode:
            message = "mode mismatch: %s != %s"
            raise IOError(message % (fname.mode, mode))
        return fname
    else:
        message = "unknown data type (expected string or filehandle): %s"
        raise SystemExit(message % type(fname))


def readtxt(fname, path="", flat=False):
    """Commodity function for reading text files plain or zipped.

    Read a text file line by line either plainly from a directory or a .zip or
    .jar file. Return as a list of strings or optionally flattened into a
    single string.

    BEWARE: this is NOT intended for HUGE text files as it actually reads them
    in and returns the content, not a handle to the reader itself!

    Parameters
    ----------
    fname : str
        The name of the file to read in. Can be a full or relative path if
        desired. For automatic archive handling use the 'path' parameter.
    path : str (optional)
        The directory where to look for the file. If the string has the suffix
        '.zip' or '.jar' an archive is assumed and the corresponding mechanisms
        are used to read 'fname' from within this archive.
    flat : bool (optional)
        Used to request a flattened string instead of a list of strings.

    Returns
    -------
    txt : str or list(str)

    Example
    -------
    >>> readtxt('foo', '/tmp/archive.zip', flat=True)
    ... # doctest: +SKIP
    """
    zipread = None
    suffix = splitext(path)[1].lower()
    if (suffix == ".zip") or (suffix == ".jar"):
        # ZipFile only works as a context manager from Python 2.7 on
        # tag:python25
        zipread = zipfile.ZipFile(path, "r")
        fin = zipread.open(fname)
    else:
        fin = open(join(path, fname), "r")
    txt = fin.readlines()  # returns file as a list, one entry per line
    try:
        txt = [x.decode("utf-8") for x in txt]
    except AttributeError:
        pass  # in Python2 decoding isn't necessary
    if flat:
        txt = flatten(txt)
    fin.close()
    if zipread is not None:
        zipread.close()
    return txt



"""Miscellaneous ImageJ related functions, mostly convenience wrappers."""

import csv
import glob
import os
import smtplib
import subprocess
import sys
import time

from ij import IJ  # pylint: disable-msg=import-error
from ij.plugin import Duplicator, ImageCalculator, StackWriter

from .. import pathtools
from ..log import LOG as log
from . import bioformats as bf
from . import prefs

from org.scijava.widget import WidgetStyle
from org.scijava.widget import TextWidget


def show_status(msg):
    """Update the ImageJ status bar and issue a log message.

    Parameters
    ----------
    msg : str
        The message to display in the ImageJ status bar and log.
    """
    log.info(msg)
    IJ.showStatus(msg)


def show_progress(cur, final):
    """Update ImageJ's progress bar and print the current progress to the log.

    Parameters
    ----------
    cur : int
        Current progress value.
    final : int
        Total value representing 100% completion.

    Notes
    -----
    `ij.IJ.showProgress` internally increments the given `cur` value by 1.
    """
    log.info("Progress: %s / %s (%s)", cur + 1, final, (1.0 + cur) / final)
    IJ.showProgress(cur, final)


def error_exit(msg):
    """Log an error message and exit.

    Parameters
    ----------
    msg : str
        The error message to log.
    """
    log.error(msg)
    sys.exit(msg)


def elapsed_time_since(start, end=None):
    """Generate a string with the time elapsed between two timepoints.

    Parameters
    ----------
    start : float
        The start time, typically obtained via `time.time()`.
    end : float, optional
        The end time. If not given, the current time is used.

    Returns
    -------
    str
        The elapsed time formatted as `HH:MM:SS.ss`.

    """
    if not end:
        end = time.time()

    hours, rem = divmod(end - start, 3600)
    minutes, seconds = divmod(rem, 60)
    return "{:0>2}:{:0>2}:{:05.2f}".format(int(hours), int(minutes), seconds)


def percentage(part, whole):
    """Calculate the percentage of a value based on total.

    Parameters
    ----------
    part : float
        The portion value of a total.
    whole : float
        The total value.

    Returns
    -------
    float
        The percentage value.
    """
    return 100 * float(part) / float(whole)


def calculate_mean_and_stdv(values_list, round_decimals=0):
    """Calculate mean and standard deviation from a list of floats.

    Parameters
    ----------
    values_list : list of int,float
        List containing numbers.
    round_decimals : int, optional
        Rounding decimal to use for the result, by default 0

    Returns
    -------
    tuple of (float, float)
        Mean and standard deviation of the input list.

    Notes
    -----
    Returns (0, 0) when:
        - The input list is empty.
        - After filtering out None values, no elements remain.
    """

    filtered_list = [x for x in values_list if x is not None]

    if not filtered_list:
        return 0, 0

    mean = round(sum(filtered_list) / len(filtered_list), round_decimals)
    variance = sum((x - mean) ** 2 for x in filtered_list) / len(filtered_list)
    std_dev = round(variance**0.5, round_decimals)

    return mean, std_dev


def find_focus(imp):
    """Find the slice of a stack that is the best focused one.

    First, calculate the variance of the pixel values in each slice. The slice
    with the highest variance is considered the best focused as this typically
    indicates more contrast and sharpness.

    Parameters
    ----------
    imp : ij.ImagePlus
        A single-channel ImagePlus stack.

    Returns
    -------
    int
        The slice number of the best focused slice.

    Raises
    ------
    SystemExit
        If the image has more than one channel.

    Notes
    -----
    Currently only single-channel stacks are supported.
    """

    imp_dimensions = imp.getDimensions()

    # Check if more than 1 channel
    # FUTURE Could be improved for multi channel
    if imp_dimensions[2] != 1:
        sys.exit("Image has more than one channel, please reduce dimensionality")

    # Loop through each time point
    for plane in range(1, imp_dimensions[4] + 1):
        focused_slice = 0
        norm_var = 0
        imp.setT(plane)
        # Loop through each z plane
        for current_z in range(1, imp_dimensions[3] + 1):
            imp.setZ(current_z)
            pix_array = imp.getProcessor().getPixels()
            mean = (sum(pix_array)) / len(pix_array)
            pix_array = [(x - mean) * (x - mean) for x in pix_array]
            # pix_array = pix_array*pix_array

            sumpix_array = sum(pix_array)
            var = sumpix_array / (imp_dimensions[0] * imp_dimensions[1] * mean)

            if var > norm_var:
                norm_var = var
                focused_slice = current_z

    return focused_slice


def send_notification_email(
    job_name, recipient, filename, total_execution_time, subject="", message=""
):
    """Send an email notification with optional details of the processed job.

    Retrieve the sender email and SMTP server settings from ImageJ's preferences
    and use them to send an email notification with job details.

    Parameters
    ----------
    job_name : string
        Job name to display in the email.
    recipient : string
        Recipient's email address.
    filename : string
        The name of the file to be passed in the email.
    total_execution_time : str
        The time it took to process the file in the format [HH:MM:SS:ss].
    subject : string, optional
        Subject of the email, by default says job finished.
    message : string, optional
        Message to be included in the email, by default says job processed.

    Notes
    -----
    - The function requires two preferences to be set in `~/.imagej/IJ_Prefs.txt`:
      - `.imcf.sender_email`: the sender's email address
      - `.imcf.smtpserver`: the SMTP server address
    - If these preferences are not set or if required parameters are missing,
      the function logs a message and exits without sending an email.
    - In case of an SMTP error, the function logs a warning.
    """

    # Retrieve sender email and SMTP server from preferences
    # NOTE: the leading dot "." has to be omitted in the `Prefs.get()` call,
    # despite being present in the `IJ_Prefs.txt` file!
    sender = prefs.Prefs.get("imcf.sender_email", "").strip()
    server = prefs.Prefs.get("imcf.smtpserver", "").strip()

    # Ensure the sender and server are configured from Prefs
    if not sender:
        log.info("[.imcf.sender_email] is not configured in '~/.imagej/IJ_Prefs.txt'.")
        return
    if not server:
        log.info("[.imcf.smtpserver] is not configured in '~/.imagej/IJ_Prefs.txt'.")
        return

    log.debug("Using SMTP server [%s].", server)

    # Ensure the recipient is provided
    if not recipient.strip():
        log.info("Recipient email is required, not sending email notification.")
        return

    # Form the email subject and body
    if subject == "":
        subject = "Your {0} job has finished".format(job_name)
    else:
        subject = subject

    if message == "":
        body = (
            "Dear recipient,\n\n"
            "This is an automated message.\n"
            "Your workflow '{0}' has been processed "
            "({1} [HH:MM:SS:ss]).\n\n"
            "Kind regards.\n"
        ).format(filename, total_execution_time)
    else:
        body = message

    # Form the complete message
    message = ("From: {0}\nTo: {1}\nSubject: {2}\n\n{3}").format(
        sender, recipient, subject, body
    )

    # Try sending the email, print error message if it wasn't possible
    try:
        smtpObj = smtplib.SMTP(server)
        smtpObj.sendmail(sender, recipient, message)
        log.debug("Successfully sent email to <%s>.", recipient)
    except smtplib.SMTPException as err:
        log.warning("Error: Unable to send email: %s", err)
    return


def progressbar(progress, total, line_number, prefix=""):
    """Progress bar for the IJ log window.

    Show a progress bar in the log window of Fiji at a specific line independent
    of the main Fiji progress bar.

    Parameters
    ----------
    progress : int
        Current step of the loop.
    total : int
        Total number of steps for the loop.
    line_number : int
        Number of the line to be updated.
    prefix : str, optional
        Text to use before the progress bar, by default an empty string.
    """

    size = 20
    x = int(size * progress / total)
    IJ.log(
        "\\Update%i:%s[%s%s] %i/%i\r"
        % (
            line_number,
            timed_log(prefix, as_string=True),
            "#" * x,
            "." * (size - x),
            progress,
            total,
        )
    )


def timed_log(message, as_string=False):
    """Print a message to the ImageJ log window, prefixed with a timestamp.

    If `as_string` is set to True, nothing will be printed to the log window,
    instead the formatted log message will be returned as a string.

    Parameters
    ----------
    message : str
        Message to print
    as_string : bool, optional
        Flag to request the formatted string to be returned instead of printing
        it to the log. By default False.
    """
    formatted = time.strftime("%H:%M:%S", time.localtime()) + ": " + message + " "
    if as_string:
        return formatted
    IJ.log(formatted)


def get_free_memory():
    """Get the free memory that is available to ImageJ.

    Returns
    -------
    free_memory : int
        The free memory in bytes.
    """
    max_memory = int(IJ.maxMemory())
    used_memory = int(IJ.currentMemory())
    free_memory = max_memory - used_memory

    return free_memory


def setup_clean_ij_environment(rm=None, rt=None):  # pylint: disable-msg=unused-argument
    """Set up a clean and defined ImageJ environment.

    This funtion clears the active results table, the ROI manager, and the log.
    Additionally, it closes all open images and resets the ImageJ options,
    performing a [*Fresh Start*][fresh_start].

    Parameters
    ----------
    rm : RoiManager, optional
        Will be ignored (kept for keeping API compatibility).
    rt : ResultsTable, optional
        Will be ignored (kept for keeping API compatibility).

    Notes
    -----
    "Fresh Start" is described in the [ImageJ release notes][ij_relnotes],
    following a [suggestion by Robert Haase][fresh_start].

    [ij_relnotes]: https://imagej.nih.gov/ij/notes.html
    [fresh_start]: https://forum.image.sc/t/43102
    """

    IJ.run("Fresh Start", "")
    IJ.log("\\Clear")

    prefs.fix_ij_options()


def sanitize_image_title(imp):
    """Remove special chars and various suffixes from the title of an ImagePlus.

    Parameters
    ----------
    imp : ImagePlus
        The ImagePlus to be renamed.

    Notes
    -----
    The function removes the full path of the image file (if present), retaining
    only the base filename using `os.path.basename()`.
    """
    # sometimes (unclear when) the title contains the full path, therefore we
    # simply call `os.path.basename()` on it to remove all up to the last "/":
    image_title = os.path.basename(imp.getTitle())
    image_title = image_title.replace(".czi", "")
    image_title = image_title.replace(" ", "_")
    image_title = image_title.replace("_-_", "")
    image_title = image_title.replace("__", "_")
    image_title = image_title.replace("#", "Series")

    imp.setTitle(image_title)


def subtract_images(imp1, imp2):
    """Subtract one image from the other (imp1 - imp2).

    Parameters
    ----------
    imp1: ij.ImagePlus
        The ImagePlus that is to be subtracted from.
    imp2: ij.ImagePlus
        The ImagePlus that is to be subtracted.

    Returns
    -------
    ij.ImagePlus
        The ImagePlus resulting from the subtraction.
    """
    ic = ImageCalculator()
    if imp1.getNSlices() != imp2.getNSlices():
        raise ValueError(
            "Cannot subtract images with different number of slices, "
            "please check your input data."
        )
    option = " stack" if imp1.getNSlices() > 1 else ""
    subtracted = ic.run("Subtract create" + option, imp1, imp2)
    subtracted.setCalibration(imp1.getCalibration())

    return subtracted


def close_images(list_of_imps):
    """Close all open ImagePlus objects given in a list.

    Parameters
    ----------
    list_of_imps: list(ij.ImagePlus)
        A list of open ImagePlus objects.
    """
    for imp in list_of_imps:
        imp.changes = False
        imp.close()


def get_threshold_value_from_method(imp, method, ops):
    """Get the value of a selected AutoThreshold method for the given ImagePlus.

    This is useful to figure out which threshold value will be calculated by the
    selected method for the given stack *without* actually having to apply it.

    Parameters
    ----------
    imp : ij.ImagePlus
        The image from which to get the threshold value.
    method : {'huang', 'ij1', 'intermodes', 'isoData', 'li', 'maxEntropy',
        'maxLikelihood', 'mean', 'minError', 'minimum', 'moments', 'otsu',
        'percentile', 'renyiEntropy', 'rosin', 'shanbhag', 'triangle', 'yen'}
        The AutoThreshold method to use.
    ops: ops.OpService
        The ImageJ Ops service instance, usually retrieved through a _Script
        Parameter_ at the top of the script, as follows:
        ```
        #@ OpService ops
        ```

    Returns
    -------
    int
        The threshold value chosen by the selected method.
    """
    histogram = ops.run("image.histogram", imp)
    threshold_value = ops.run("threshold.%s" % method, histogram)
    threshold_value = int(round(threshold_value.get()))

    return threshold_value


def write_ordereddict_to_csv(out_file, content):
    """Write data from a list of OrderedDicts to a CSV file.

    When performing measurements in an analysis that is e.g. looping over
    multiple files, it's useful to keep the results in `OrderedDict` objects,
    e.g. one per analyzed file / dataset. This function can be used to create a
    CSV file (or append to an existing one) from a list of `OrderedDict`s. The
    structure inside the dicts is entirely up to the calling code (i.e. it's not
    related to ImageJ's *Results* window or such), the only requirement is
    type-consistency among all the `OrderedDict`s provided to the function.

    Parameters
    ----------
    out_file : str
        Path to the output CSV file.
    content : list of OrderedDict
        List of OrderedDict objects representing the data rows to be written.
        All dictionaries must have the same keys.

    Notes
    -----
    - The CSV file will use the semicolon charachter (`;`) as delimiter.
    - When appending to an existing file, the column structure has to match. No
      sanity checking is being done on this by the function!
    - The output file is opened in binary mode for compatibility.

    Examples
    --------
    >>> from collections import OrderedDict
    >>> results = [
    ...     OrderedDict([('id', 1), ('name', 'Sample A'), ('value', 42.5)]),
    ...     OrderedDict([('id', 2), ('name', 'Sample B'), ('value', 37.2)])
    ... ]
    >>> write_ordereddict_to_csv('results.csv', results)

    The resulting CSV file will have the following content:

        id;name;value
        1;Sample A;42.5
        2;Sample B;37.2
    """

    # Check if the output file exists
    if not os.path.exists(out_file):
        # If the file does not exist, create it and write the header
        with open(out_file, "wb") as f:
            dict_writer = csv.DictWriter(f, content[0].keys(), delimiter=";")
            dict_writer.writeheader()
            dict_writer.writerows(content)
    else:
        # If the file exists, append the results
        with open(out_file, "ab") as f:
            dict_writer = csv.DictWriter(f, content[0].keys(), delimiter=";")
            dict_writer.writerows(content)


def save_image_in_format(imp, format, out_dir, series, pad_number, split_channels, suffix=""):
    """Save an ImagePlus object in the specified format.

    This function provides flexible options for saving ImageJ images in various
    formats with customizable naming conventions. It supports different
    Bio-Formats compatible formats as well as ImageJ-native formats, and can
    handle multi-channel images by either saving them as a single file or
    splitting channels into separate files.

    The function automatically creates necessary directories and uses consistent
    naming patterns with series numbers. For split channels, separate
    subdirectories are created for each channel (C1, C2, etc.).

    Parameters
    ----------
    imp : ij.ImagePlus
        ImagePlus object to save.
    format : {'ImageJ-TIF', 'ICS-1', 'ICS-2', 'OME-TIFF', 'CellH5', 'BMP'}
        Output format to use, see Notes section below for details.
    out_dir : str
        Directory path where the image(s) will be saved.
    series : int
        Series number to append to the filename.
    pad_number : int
        Number of digits to use when zero-padding the series number.
    split_channels : bool
        If True, split channels and save them individually in separate folders
        named "C1", "C2", etc. inside out_dir. If False, save all channels in a
        single file.
    suffix : str, optional
        Text to be added to the filename, by default an empty string.

    Notes
    -----
    Depending on the value of the `format` parameter, one of the following
    output formats and saving strategies will be used:
    - Bio-Formats based formats will be produced by calling `bf.export()`, note
      that these formats will preserve metadata (which is **not** the case for
      the other formats using different saving strategies):
        - `ICS-1`: Save as ICS version 1 format (a pair of `.ics` and `.ids`).
        - `ICS-2`: Save as ICS version 2 format (single `.ics` file).
        - `OME-TIFF`: Save in OME-TIFF format (`.ome.tif`).
        - `CellH5`: Save as CellH5 format (`.ch5`).
    - `ImageJ-TIF`: Save in ImageJ TIFF format (`.tif`) using `IJ.saveAs()`.
    - `BMP`: Save in BMP format using `StackWriter.save()`, producing one `.bmp`
      per slice in a subfolder named after the original image.

    Examples
    --------
    Save a multichannel image as OME-TIFF without splitting channels:

    >>> save_image_with_extension(imp, "OME-TIFF", "/output/path", 1, 3, False)
    ... # resulting file: /output/path/image_title_series_001.ome.tif

    Save with channel splitting:

    >>> save_image_with_extension(imp, "OME-TIFF", "/output/path", 1, 3, True)
    ... # resulting files: /output/path/C1/image_title_series_001.ome.tif
    ... #                  /output/path/C2/image_title_series_001.ome.tif
    """

    out_ext = {}
    out_ext["ImageJ-TIF"] = ".tif"
    out_ext["ICS-1"] = ".ids"
    out_ext["ICS-2"] = ".ics"
    out_ext["OME-TIFF"] = ".ome.tif"
    out_ext["CellH5"] = ".ch5"
    out_ext["BMP"] = ".bmp"

    imp_to_use = []
    dir_to_save = []

    if split_channels:
        for channel in range(1, imp.getNChannels() + 1):
            imp_to_use.append(
                Duplicator().run(
                    imp,
                    channel,
                    channel,
                    1,
                    imp.getNSlices(),
                    1,
                    imp.getNFrames(),
                )
            )
            dir_to_save.append(os.path.join(out_dir, "C" + str(channel)))
    else:
        imp_to_use.append(imp)
        dir_to_save.append(out_dir)

    for index, current_imp in enumerate(imp_to_use):
        basename = imp.getShortTitle()

        out_path = pathtools.join2(
            dir_to_save[index],
            basename + "_series_" + str(series).zfill(pad_number) + suffix,
        )

        if format == "ImageJ-TIF":
            pathtools.create_directory(dir_to_save[index])
            IJ.saveAs(current_imp, "Tiff", out_path + suffix + ".tif")

        elif format == "BMP":
            out_folder = os.path.join(out_dir, basename + os.path.sep)
            pathtools.create_directory(out_folder)
            StackWriter.save(current_imp, out_folder, "format=bmp")

        else:
            bf.export(current_imp, out_path + suffix + out_ext[format])

        current_imp.close()


def locate_latest_imaris(paths_to_check=None):
    """Find paths to latest installed Imaris or ImarisFileConverter version.

    Identify the full path to the most recent (as in "version number")
    ImarisFileConverter or Imaris installation folder with the latter one having
    priority. In case nothing is found, an empty string is returned.

    Parameters
    ----------
    paths_to_check: list of str, optional
        A list of paths that should be used to look for the installations, by default
        `None` which will fall back to the standard installation locations of Bitplane.

    Returns
    -------
    str
    """
    if not paths_to_check:
        paths_to_check = [
            r"C:\Program Files\Bitplane\ImarisFileConverter ",
            r"C:\Program Files\Bitplane\Imaris ",
        ]

    imaris_paths = [""]

    for check in paths_to_check:
        hits = glob.glob(check + "*")
        imaris_paths += sorted(
            hits,
            key=lambda x: float(x.replace(check, "").replace(".", "")),
        )

    return imaris_paths[-1]


def run_imarisconvert(file_path, pixel_calibration=None, output_folder=""):
    """Convert a given file to Imaris format using ImarisConvert.

    Convert the input image file to Imaris format (Imaris5) using the
    ImarisConvert utility. The function uses the latest installed Imaris
    application to perform the conversion via `subprocess.call()`.

    Parameters
    ----------
    file_path : str
        Absolute path to the input image file.
    pixel_calibration : tuple or list, optional
        Sequence of 3 values (x, y, z) representing voxel dimensions to be set
        during conversion, by default None.
    output_folder : str, optional
        Folder where the newly created IMS file will be saved. If empty (or not
        supplied), the directory of the input file will be used.

    Notes
    -----
    - The output filename is constructed by replacing extension of the input
      filename with `.ims` (e.g. `/path/to/image.czi` -> `/path/to/image.ims`).
    - If the input has an `.ids` extension (part of an ICS-1 pair), the
      corresponding `.ics` file is used instead.
    """
    # in case the given file has the suffix `.ids` (meaning it is part of an
    # ICS-1 `.ics`+`.ids` pair), point ImarisConvert to the `.ics` file instead:
    path_root, file_extension = os.path.splitext(file_path)
    if file_extension == ".ids":
        file_extension = ".ics"
        file_path = path_root + file_extension

    imaris_path = locate_latest_imaris()

    if not output_folder:
        output_folder = os.path.dirname(file_path)

    command = 'ImarisConvert.exe  -i "%s" -of Imaris5 -o "%s"' % (
        file_path,
        os.path.join(output_folder, file_path.replace(file_extension, ".ims")),
    )
    if pixel_calibration:
        command = command + " --voxelsizex %s --voxelsizey %s --voxelsizez %s" % (
            pixel_calibration[0],
            pixel_calibration[1],
            pixel_calibration[2],
        )

    log.debug("\n%s" % command)
    timed_log("Converting to Imaris5 .ims...")
    result = subprocess.call(command, shell=True, cwd=imaris_path)
    if result == 0:
        timed_log("Conversion to .ims is finished: %s" % file_path)
    else:
        timed_log("Error converting [%s]: %d" % (file_path, result))


def save_script_parameters(destination, save_file_name="script_parameters.txt"):
    """Save all Fiji script parameters to a text file.

    Parameters
    ----------
    destination : str
        Directory where the script parameters file will be saved.
    save_file_name : str, optional
        Name of the script parameters file, by default "script_parameters.txt".

    Notes
    -----
    This function records all input parameters defined in the Fiji script header
    (e.g. `#@ String`) to a text file.

    The following parameters are excluded:
    - Parameters explicitly declared with `style="password"` are ignored.
    - Runtime keys (e.g. 'SJLOG', 'COMMAND', 'RM') are also skipped.
    """
    # Get the ScriptModule object from globals made by Fiji
    module = globals().get("org.scijava.script.ScriptModule")
    if module is None:
        timed_log("No ScriptModule found - skipping saving script parameters.")
        return

    destination = str(destination)
    out_path = os.path.join(destination, save_file_name)

    # Access script metadata and inputs
    script_info = module.getInfo()
    inputs = module.getInputs()

    # Keys to skip explicitly
    skip_keys = ["USERNAME", "SJLOG", "COMMAND", "RM"]

    with open(out_path, "w") as f:
        for item in script_info.inputs():
            key = item.getName()

            # Skip if any keys are in the skip list
            if any(skip in key.upper() for skip in skip_keys):
                continue

            # Skip if parameter is declared with password style
            if WidgetStyle.isStyle(item, TextWidget.PASSWORD_STYLE):
                continue

            if inputs.containsKey(key):
                val = inputs.get(key)
                f.write("%s: %s\n" % (key, str(val)))

    timed_log("Saved script parameters to: %s" % out_path)



"""BigDataViewer related functions.

Mostly convenience wrappers with simplified calls and default values.
"""

# Some function names just need to be longer than 30 chars:
# pylint: disable-msg=invalid-name

# The attribute count is not really our choice:
# pylint: disable-msg=too-many-instance-attributes

import os
import shutil
import sys

from ch.epfl.biop.scijava.command.spimdata import (
    FuseBigStitcherDatasetIntoOMETiffCommand,
)
from ij import IJ
from java.io import File, FileInputStream, InputStreamReader
from javax.xml.parsers import DocumentBuilderFactory
from org.xml.sax import InputSource

from .. import pathtools
from ..log import LOG as log

# internal template strings used in string formatting (note: the `"""@private"""`
# pseudo-decorator is there to instruct [pdoc] to omit those variables when generating
# API documentation):
SINGLE = "[Single %s (Select from List)]"
"""@private"""
MULTIPLE = "[Multiple %ss (Select from List)]"
"""@private"""
RANGE = "[Range of %ss (Specify by Name)]"
"""@private"""
SINGLE_FILE = "[NO (one %s)]"
"""@private"""
MULTI_SINGLE_FILE = "[YES (all %ss in one file)]"
"""@private"""
MULTI_MULTI_FILE = "[YES (one file per %s)]"
"""@private"""


class ProcessingOptions(object):
    """Helper to store processing options and generate parameter strings.

    Example
    -------
    NOTE: for readability reasons the output has been split into multiple lines
    even though the formatters are returning a single-line string.

    >>> opts = ProcessingOptions()
    >>> opts.process_channel(2)
    >>> opts.reference_tile(1)
    >>> opts.treat_timepoints("compare")

    >>> opts.fmt_acitt_options()
    ... process_angle=[All angles]
    ... process_channel=[Single channel (Select from List)]
    ... process_illumination=[All illuminations]
    ... process_tile=[All tiles]
    ... process_timepoint=[All Timepoints]

    >>> opts.fmt_acitt_selectors()
    ... processing_channel=[channel 1]

    >>> opts.fmt_use_acitt()
    ... channels=[Average Channels]
    ... illuminations=[Average Illuminations]
    ... tiles=[use Tile 1]

    >>> opts.fmt_how_to_treat()
    ... how_to_treat_angles=[treat individually]
    ... how_to_treat_channels=group
    ... how_to_treat_illuminations=group
    ... how_to_treat_tiles=group
    ... how_to_treat_timepoints=compare
    """

    def __init__(self):
        self._angle_processing_option = "[All angles]"
        self._angle_select = ""

        self._channel_processing_option = "[All channels]"
        self._channel_select = ""

        self._illumination_processing_option = "[All illuminations]"
        self._illumination_select = ""

        self._tile_processing_option = "[All tiles]"
        self._tile_select = ""

        self._timepoint_processing_option = "[All Timepoints]"
        self._timepoint_select = ""

        # by default `angles` is empty as the "sane" default value for
        # "treat_angles" is "[treat individually]"
        self._use_angle = ""
        # all other "use" options are set to averaging by default:
        self._use_channel = "channels=[Average Channels]"
        self._use_illumination = "illuminations=[Average Illuminations]"
        self._use_tile = "tiles=[Average Tiles]"
        self._use_timepoint = "timepoints=[Average Timepoints]"

        # 'treat_*' values are: "group", "compare" or "[treat individually]"
        self._treat_angles = "[treat individually]"
        self._treat_channels = "group"
        self._treat_illuminations = "group"
        self._treat_tiles = "compare"
        self._treat_timepoints = "[treat individually]"

    ### reference-X methods

    def reference_angle(self, value):
        """Set the reference angle when using *Expert Grouping Options*.

        Select the angle(s) to use for the operation, by default empty (`""`).

        NOTE: this value will be used to render `angles=[use Angle VALUE]` when
        calling the `fmt_use_acitt()` method.

        Parameters
        ----------
        value : str
            The tile to use for the grouping.
        """
        self._use_angle = "angles=[use Angle %s]" % str(value)
        log.debug("New reference angle setting: %s", self._use_angle)

    def reference_channel(self, value):
        """Set the reference channel when using *Expert Grouping Options*.

        Select the channel(s) to use for the operation, by default the averaging
        mode will be used (`channels=[Average Channels]`).

        NOTE: this value will be used to render `channels=[use Channel VALUE]`
        when calling the `fmt_use_acitt()` method.

        Parameters
        ----------
        value : int or int-like
            The channel number to use for the grouping.
        """
        # channel = int(value) - 1  # will raise a ValueError if cast fails
        self._use_channel = "channels=[use Channel %s]" % int(value)
        log.debug("New reference channel setting: %s", self._use_channel)

    def reference_illumination(self, value):
        """Set the reference illumination when using *Expert Grouping Options*.

        Select the illumination(s) to use for the operation, by default the
        averaging mode will be used (`illuminations=[Average Illuminations]`).

        NOTE: this value will be used to render `illuminations=[use Illumination
        VALUE]` when calling the `fmt_use_acitt()` method.

        Parameters
        ----------
        value : int or int-like
            The illumination number to use for the grouping.
        """
        self._use_illumination = "illuminations=[use Illumination %s]" % value
        log.debug(
            "New reference illumination setting: %s",
            self._use_illumination,
        )

    def reference_tile(self, value):
        """Set the reference tile when using *Expert Grouping Options*.

        Select the tile(s) to use for the operation, by default the averaging
        mode will be used (`tiles=[Average Tiles]`).

        NOTE: this value will be used to render `tiles=[use Tile VALUE]` when
        calling the `fmt_use_acitt()` method.

        Parameters
        ----------
        value : int
            The tile number to use for the grouping.
        """
        self._use_tile = "tiles=[use Tile %s]" % str(value)
        log.debug("New reference tile setting: %s", self._use_tile)

    def reference_timepoint(self, value):
        """Set the reference timepoint when using *Expert Grouping Options*.

        Select the timepoint(s) to use for the operation, by default the
        averaging mode will be used (`timepoints=[Average Timepoints]`).

        NOTE: this value will be used to render `timepoints=[use Timepoint
        VALUE]` when calling the `fmt_use_acitt()` method.

        Parameters
        ----------
        value : int or int-like
            The timepoint number to use for the grouping.
        """
        self._use_timepoint = "timepoints=[use Timepoint %s]" % value
        log.debug("New reference timepoint setting: %s", self._use_timepoint)

    ### process-X methods

    def process_angle(self, value, range_end=None):
        """Set the processing option for angles.

        Update the angle processing option and selection depending on input.
        If the range_end is not None, it is considered as a range.

        Parameters
        ----------
        value : str, int, list of int or list of str
            The angle(s) to use for processing, either a single value or a list.
        range_end : int, optional
            Contains the end of the range, by default None.

        Notes
        -----
        Previous function name : angle_select().
        """

        selection = check_processing_input(value, range_end)
        processing_option, dimension_select = get_processing_settings(
            "angle", selection, value, range_end
        )

        self._angle_processing_option = processing_option
        self._angle_select = dimension_select

    def process_channel(self, value, range_end=None):
        """Set the processing option for channels.

        Update the channel processing option and selection depending on input.
        If the range_end is not None, it is considered as a range.

        Parameters
        ----------
        value : str, int, list of int or list of str
            The channel(s) to use for processing, a single value or a list.
        range_end : int, optional
            Contains the end of the range, by default None.

        Notes
        -----
        Previous function name : channel_select().
        """

        selection = check_processing_input(value, range_end)
        processing_option, dimension_select = get_processing_settings(
            "channel", selection, value, range_end
        )

        self._channel_processing_option = processing_option
        self._channel_select = dimension_select

    def process_illumination(self, value, range_end=None):
        """Set the processing option for illuminations.

        Update the illumination processing option and selection depending on
        input. If the range_end is not None, it is considered as a range.

        Parameters
        ----------
        value : str, int, list of int or list of str
            The illumination(s) to use for processing, a single value or a list.
        range_end : int, optional
            Contains the end of the range, by default None.

        Notes
        -----
        Previous function name : illumination_select().
        """

        selection = check_processing_input(value, range_end)
        processing_option, dimension_select = get_processing_settings(
            "illumination", selection, value, range_end
        )

        self._illumination_processing_option = processing_option
        self._illumination_select = dimension_select

    def process_tile(self, value, range_end=None):
        """Set the processing option for tiles.

        Update the tile processing option and selection depending on input.
        If the range_end is not None, it is considered as a range.

        Parameters
        ----------
        value : str, int, list of int or list of str
            The tile(s) to use for processing, a single value or a list.
        range_end : int, optional
            Contains the end of the range, by default None.

        Notes
        -----
        Previous function name : tile_select().
        """

        selection = check_processing_input(value, range_end)
        processing_option, dimension_select = get_processing_settings(
            "tile", selection, value, range_end
        )

        self._tile_processing_option = processing_option
        self._tile_select = dimension_select

    def process_timepoint(self, value, range_end=None):
        """Set the processing option for timepoints.

        Update the timepoint processing option and selection depending on input.
        If the range_end is not None, it is considered as a range.

        Parameters
        ----------
        value : str, int, list of int or list of str
            The timepoint(s) to use for processing, a single value or a list.
        range_end : int, optional
            Contains the end of the range, by default None.

        Notes
        -----
        Previous function name : timepoint_select().
        """

        selection = check_processing_input(value, range_end)
        processing_option, dimension_select = get_processing_settings(
            "timepoint", selection, value, range_end
        )

        self._timepoint_processing_option = processing_option
        self._timepoint_select = dimension_select

    ### treat-X methods

    def treat_angles(self, value):
        """Set the value for the `how_to_treat_angles` option.

        If the value is set to `group` also the `reference_angle` setting will
        be adjusted to `angles=[Average Angles]`.

        The default setting is `[treat individually]`.

        Parameters
        ----------
        value : str
            One of `group`, `compare` or `[treat individually]`.
        """
        self._treat_angles = value
        log.debug("New 'treat_angles' setting: %s", value)
        if value == "group":
            self._use_angle = "angles=[Average Angles]"
            log.debug("New 'use_angle' setting: %s", self._use_angle)

    def treat_channels(self, value):
        """Set the value for the `how_to_treat_channels` option.

        The default setting is `group`.

        Parameters
        ----------
        value : str
            One of `group`, `compare` or `[treat individually]`.
        """
        self._treat_channels = value
        log.debug("New 'treat_channels' setting: %s", value)

    def treat_illuminations(self, value):
        """Set the value for the `how_to_treat_illuminations` option.

        The default setting is `group`.

        Parameters
        ----------
        value : str
            One of `group`, `compare` or `[treat individually]`.
        """
        self._treat_illuminations = value
        log.debug("New 'treat_illuminations' setting: %s", value)

    def treat_tiles(self, value):
        """Set the value for the `how_to_treat_tiles` option.

        The default setting is `compare`.

        Parameters
        ----------
        value : str
            One of `group`, `compare` or `[treat individually]`.
        """
        self._treat_tiles = value
        log.debug("New 'treat_tiles' setting: %s", value)

    def treat_timepoints(self, value):
        """Set the value for the `how_to_treat_timepoints` option.

        The default setting is `[treat individually]`.

        Parameters
        ----------
        value : str
            One of `group`, `compare` or `[treat individually]`.
        """
        self._treat_timepoints = value
        log.debug("New 'treat_timepoints' setting: %s", value)

    ### formatter methods

    def fmt_acitt_options(self, input="process"):
        """Format Angle / Channel / Illumination / Tile / Timepoint options.

        Build a string providing the `process_angle`, `process_channel`,
        `process_illumination`, `process_tile` and `process_timepoint` options
        that can be used in a BDV-related `IJ.run` call.

        Returns
        -------
        str
        """
        input_type = ["process", "resave"]
        if input not in input_type:
            raise ValueError("Invalid input type, expected one of: %s" % input_type)
        parameters = [
            input + "_angle=" + self._angle_processing_option,
            input + "_channel=" + self._channel_processing_option,
            input + "_illumination=" + self._illumination_processing_option,
            input + "_tile=" + self._tile_processing_option,
            input + "_timepoint=" + self._timepoint_processing_option,
        ]
        parameter_string = " ".join(parameters).strip()
        log.debug("Formatted 'process_X' options: <%s>", parameter_string)
        return parameter_string + " "

    def fmt_acitt_selectors(self):
        """Format Angle / Channel / Illumination / Tile / Timepoint selectors.

        Build a string providing the `angle_select`, `channel_select`,
        `illumination_select`, `tile_select` and `timepoint_select` options that
        can be used in a BDV-related `IJ.run` call. In case no selectors have
        been chosen, nothing but a single space will be returned.

        Returns
        -------
        str
            The formatted selector string. Will be a single white-space in case
            no selectors have been configured for the object.
        """
        parameters = [
            self._angle_select if self._angle_select else "",
            self._channel_select if self._channel_select else "",
            self._illumination_select if self._illumination_select else "",
            self._tile_select if self._tile_select else "",
            self._timepoint_select if self._timepoint_select else "",
        ]
        parameter_string = " ".join(parameters).strip()
        log.debug("Formatted 'processing_X' selectors: <%s>", parameter_string)
        return parameter_string + " "

    def fmt_how_to_treat(self):
        """Format a parameter string with all `how_to_treat_` options.

        Returns
        -------
        str
        """
        parameters = [
            "how_to_treat_angles=" + self._treat_angles,
            "how_to_treat_channels=" + self._treat_channels,
            "how_to_treat_illuminations=" + self._treat_illuminations,
            "how_to_treat_tiles=" + self._treat_tiles,
            "how_to_treat_timepoints=" + self._treat_timepoints,
        ]
        parameter_string = " ".join(parameters).strip()
        log.debug("Formatted 'how_to_treat_X' options: <%s>", parameter_string)
        return parameter_string + " "

    def fmt_use_acitt(self):
        """Format expert grouping options, e.g. `channels=[use Channel 2]`.

        Generate a parameter string using the configured expert grouping options
        for ACITT. Please note that this may be an empty string (`""`).

        Returns
        -------
        str
        """
        parameters = [
            self._use_angle if self._treat_angles == "group" else "",
            self._use_channel if self._treat_channels == "group" else "",
            self._use_illumination if self._treat_illuminations == "group" else "",
            self._use_tile if self._treat_tiles == "group" else "",
            self._use_timepoint if self._treat_timepoints == "group" else "",
        ]
        parameter_string = " ".join(parameters).strip()
        log.debug(
            "Formatted expert grouping 'use' options: <%s>",
            parameter_string,
        )
        return parameter_string + " "


class DefinitionOptions(object):
    """Helper to store definition options and generate parameters strings.

    Example
    -------
    NOTE: for readability reasons the output has been split into multiple lines
    even though the formatters are returning a single-line string.

    >>> opts = DefinitionOptions()
    >>> opts.set_angle_definition("single")
    >>> opts.set_channel_definition("multi_single")

    >>> opts.fmt_acitt_options()
    ... multiple_angles=[NO (one angle)]
    ... multiple_channels=[YES (all channels in one file)]
    ... multiple_illuminations_directions=[NO (one illumination direction)]
    ... multiple_tiles=[YES (all tiles in one file)]
    ... multiple_timepoints=[NO (one time-point)]
    """

    def __init__(self):
        self._angle_definition = SINGLE_FILE % "angle"
        self._channel_definition = MULTI_SINGLE_FILE % "channel"
        self._illumination_definition = SINGLE_FILE % "illumination direction"
        self._tile_definition = MULTI_MULTI_FILE % "tile"
        self._timepoint_definition = SINGLE_FILE % "time-point"

    def check_definition_option(self, value):
        """Check if the value is a valid definition option.

        Parameters
        ----------
        value : str
            Entered value by the user.

        Returns
        -------
        dict(str, str): dictionary containing the correct string definition.
        """
        valid = ["single", "multi_single", "multi_multi"]
        if value not in valid:
            raise ValueError("Value must be one of: %s" % valid)

        return {
            "single": SINGLE_FILE,
            "multi_single": MULTI_SINGLE_FILE,
            "multi_multi": MULTI_MULTI_FILE,
        }

    def check_definition_option_ang_ill(self, value):
        """Check if the value is a valid definition option.

        This is needed for angles and illuminations because support is not
        available for multiple angles and illuminations in a single file.

        Parameters
        ----------
        value : str
            Entered value by the user.

        Returns
        -------
        dict(str, str): dictionary containing the correct string definition.
        """
        valid = ["single", "multi_multi"]
        if value not in valid:
            raise ValueError(
                (
                    "Value must be one of: %s. Support for 'multi_single' is "
                    "not available for angles and illuminations."
                )
                % valid
            )

        return {
            "single": SINGLE_FILE,
            "multi_multi": MULTI_MULTI_FILE,
        }

    def set_angle_definition(self, value):
        """Set the value for the angle definition.

        Parameters
        ----------
        value : str
            One of `single` or `multi_multi`.
        """
        choices = self.check_definition_option_ang_ill(value)
        self._angle_definition = choices[value] % "angle"
        log.debug("New 'angle_definition' setting: %s", self._angle_definition)

    def set_channel_definition(self, value):
        """Set the value for the channel definition.

        Parameters
        ----------
        value : str
            One of `single`, `multi_single` or `multi_multi`.
        """
        choices = self.check_definition_option(value)
        self._channel_definition = choices[value] % "channel"
        log.debug(
            "New 'channel_definition' setting: %s",
            self._channel_definition,
        )

    def set_illumination_definition(self, value):
        """Set the value for the illumination definition.

        Parameters
        ----------
        value : str
            One of `single`, `multi_single` or `multi_multi`.
        """
        choices = self.check_definition_option_ang_ill(value)
        self._illumination_definition = choices[value] % "illumination direction"
        log.debug(
            "New 'illumination_definition' setting: %s",
            self._illumination_definition,
        )

    def set_tile_definition(self, value):
        """Set the value for the tile_definition.

        Parameters
        ----------
        value : str
            One of `single`, `multi_single` or `multi_multi`.
        """
        choices = self.check_definition_option(value)
        self._tile_definition = choices[value] % "tile"
        log.debug("New 'tile_definition' setting: %s", self._tile_definition)

    def set_timepoint_definition(self, value):
        """Set the value for the time_point_definition.

        Parameters
        ----------
        value : str
            One of `single`, `multi_single` or `multi_multi`.
        """
        choices = self.check_definition_option(value)
        self._timepoint_definition = choices[value] % "time-point"
        log.debug(
            "New 'timepoint_definition' setting: %s",
            self._timepoint_definition,
        )

    def fmt_acitt_options(self):
        """Format Angle / Channel / Illumination / Tile / Timepoint options.

        Build a string providing the `multiple_angles`, `multiple_channels`,
        `multiple_illuminations_directions`, `multiple_tiles` and
        `multiple_timepoints` options that can be used in a BDV-related `IJ.run`
        call.

        Returns
        -------
        str
        """
        parameters = [
            "multiple_angles=" + self._angle_definition,
            "multiple_channels=" + self._channel_definition,
            "multiple_illuminations_directions=" + self._illumination_definition,
            "multiple_tiles=" + self._tile_definition,
            "multiple_timepoints=" + self._timepoint_definition,
        ]
        parameter_string = " ".join(parameters).strip()
        log.debug("Formatted 'multiple_X' options: <%s>", parameter_string)
        return parameter_string + " "


def check_processing_input(value, range_end):
    """Sanitize and clarifies the acitt input selection.

    Validate the input by checking the type and returning the expected output.

    Parameters
    ----------
    value : str, int, list of int or list of str
        Contains the list of input dimensions, the first input dimension of a
        range or a single channel.
    range_end : int or None
        Contains the end of the range if need be.

    Returns
    -------
    str
        Returns the type of selection: single, multiple or range
    """
    if type(value) is not list:
        value = [value]
    # Check if all the elements of the value list are of the same type
    if not all(isinstance(x, type(value[0])) for x in value):
        raise TypeError("Invalid input, all values must be of the same type.")
    if type(range_end) is int:
        if type(value[0]) is not int:
            raise TypeError("Range start needs to be an int.")
        elif len(value) != 1:
            raise ValueError("Range start needs to be single number.")
        else:
            return "range"
    elif len(value) == 1:
        return "single"
    else:
        return "multiple"


def get_processing_settings(dimension, selection, value, range_end):
    """Generate processing strings for selected dimension and processing mode.

    Generate the processing option and dimension selection strings that
    correspond to the selected processing mode and the given dimension
    selection.

    Parameters
    ----------
    dimension : {`angle`, `channel`, `illumination`, `tile`, `timepoint`}
        The dimension selection to use.
    selection : {`single`, `multiple`, `range`}
        The *selector* name ("processing mode"), used to derive how the
        generated string needs to be assembled according to the given dimension
        and value / range settings.
    value : str, int, list of int or list of str
        The list of input dimensions, the first input dimension of a range or a
        single dimension value in case `selection == "single"` (e.g. for
        selecting a single channel).
    range_end : int or None
        Contains the end of the range if need be.

    Returns
    -------
    tuple of str
        processing_option, dimension_select
    """
    processing_option = dimension_select = ""

    # Validate inputs according to the function docstring
    valid_dimensions = ("angle", "channel", "illumination", "tile", "timepoint")
    if dimension not in valid_dimensions:
        raise ValueError(
            "Invalid dimension '%s', expected one of: %s"
            % (dimension, ", ".join(valid_dimensions))
        )

    valid_selections = ("single", "multiple", "range")
    if selection not in valid_selections:
        raise ValueError(
            "Invalid selection '%s', expected one of: %s"
            % (selection, ", ".join(valid_selections))
        )

    if selection == "single":
        processing_option = SINGLE % dimension
        dimension_select = "processing_" + dimension + "=[" + dimension + " %s]" % value

    if selection == "multiple":
        processing_option = MULTIPLE % dimension
        dimension_list = ""
        for dimension_name in value:
            dimension_list += dimension + "_%s " % dimension_name
        dimension_select = dimension_list.rstrip()

    if selection == "range":
        processing_option = RANGE % dimension
        dimension_select = (
            "process_following_"
            + dimension
            + "s=%s-%s"
            % (
                value,
                range_end,
            )
        )

    return processing_option, dimension_select


def backup_xml_files(source_directory, subfolder_name):
    """Create a backup of BDV-XML files inside a subfolder of `xml-backup`.

    Copies all `.xml` and `.xml~` files to a subfolder with the given name
    inside a folder called `xml-backup` in the source directory. Uses the
    `shutil.copy2()` command, which will overwrite existing files.

    Parameters
    ----------
    source_directory : str
        Full path to the directory containing the xml files.
    subfolder_name : str
        The name of the subfolder that will be used inside `xml-backup`. Will be
        created if necessary.
    """
    xml_backup_directory = os.path.join(source_directory, "xml-backup")
    pathtools.create_directory(xml_backup_directory)
    backup_subfolder = xml_backup_directory + "/%s" % (subfolder_name)
    pathtools.create_directory(backup_subfolder)
    all_xml_files = pathtools.listdir_matching(source_directory, ".*\\.xml", regex=True)
    os.chdir(source_directory)
    for xml_file in all_xml_files:
        shutil.copy2(xml_file, backup_subfolder)


def define_dataset_auto(
    project_filename,
    file_path,
    bf_series_type,
    dataset_save_path=None,
    timepoints_per_partition=1,
    resave="Re-save as multiresolution HDF5",
    subsampling_factors=None,
    hdf5_chunk_sizes=None,
):
    """Define a dataset using the Autoloader or Multi-View loader.

    If the series is tiles, will run "Define Dataset...", otherwise will run
    "Define Multi-View Dataset...".

    Parameters
    ----------
    project_filename : str
        Name of the project (without an `.xml` extension).
    file_path : str
        Path to the file, can be the first `.czi` or a regex to match all files
        with an extension.
    dataset_save_path : str
        Output path for the `.xml`.
    bf_series_type : {`Angles`,`Tiles`}
        Defines how Bio-Formats interprets the series.
    timepoints_per_partition : int, optional
        Split the output dataset by timepoints. Use `0` for no split, resulting
        in a single HDF5 file containing all timepoints. By default `1`,
        resulting in a HDF5 per timepoints.
    resave : str, optional
        Allow the function to either re-save the images or simply create a
        merged xml. Use `Load raw data` to avoid re-saving, by default `Re-save
        as multiresolution HDF5` which will resave the input data.
    subsampling_factors : str, optional
        Specify subsampling factors explicitly, for example:
        `[{ {1,1,1}, {2,2,1}, {4,4,2}, {8,8,4} }]`.
    hdf5_chunk_sizes : str, optional
        Specify hdf5_chunk_sizes factors explicitly, for example
        `[{ {32,16,8}, {16,16,16}, {16,16,16}, {16,16,16} }]`.
    """

    file_info = pathtools.parse_path(file_path)

    project_filename = project_filename.replace(" ", "_")
    result_folder = pathtools.join2(file_info["path"], project_filename)

    if not os.path.exists(result_folder):
        os.makedirs(result_folder)

    if not dataset_save_path:
        dataset_save_path = result_folder
    if subsampling_factors:
        subsampling_factors = (
            "manual_mipmap_setup subsampling_factors=" + subsampling_factors + " "
        )
    else:
        subsampling_factors = ""
    if hdf5_chunk_sizes:
        hdf5_chunk_sizes = "hdf5_chunk_sizes=" + hdf5_chunk_sizes + " "
    else:
        hdf5_chunk_sizes = ""

    if bf_series_type == "Angles":
        angle_rotation = "apply_angle_rotation "
    else:
        angle_rotation = ""

    options = (
        "define_dataset=[Automatic Loader (Bioformats based)]"
        + " "
        + "project_filename=["
        + project_filename
        + ".xml"
        + "] "
        + "path=["
        + file_info["full"]
        + "] "
        + "exclude=10 "
        + "bioformats_series_are?="
        + bf_series_type
        + " "
        + "move_tiles_to_grid_(per_angle)?=["
        + "Do not move Tiles to Grid (use Metadata if available)] "
        + "how_to_store_input_images=["
        + resave
        + "] "
        + "load_raw_data_virtually"
        + " "
        + "metadata_save_path=["
        + dataset_save_path
        + "] "
        + "image_data_save_path=["
        + dataset_save_path
        + "] "
        + "check_stack_sizes "
        + angle_rotation
        + subsampling_factors
        + hdf5_chunk_sizes
        + "split_hdf5 "
        + "timepoints_per_partition="
        + str(timepoints_per_partition)
        + " "
        + "setups_per_partition=0 "
        + "use_deflate_compression "
    )

    log.debug(options)

    IJ.run("Define Multi-View Dataset", str(options))


def define_dataset_manual(
    project_filename,
    source_directory,
    image_file_pattern,
    dataset_organisation,
    definition_opts=None,
    list_files=None,
):
    """Run "Define Multi-View Dataset" using the "Manual Loader" option.

    Parameters
    ----------
    project_filename : str
        Name of the project (without an `.xml` extension).
    source_directory : str
        Path to the folder containing the file(s).
    image_file_pattern : str
        Regular expression corresponding to the names of your files and how to
        read the different dimensions.
    dataset_organisation : str
        Organisation of the dataset and the dimensions to process. Allows for
        defining the range(s) of interest for the different dimensions, for
        example: `timepoints_=%s-%s channels_=0-%s tiles_=%s-%s`.
    definition_opts : dict
        Dictionary containing the details about the file repartitions.
    list_files : list of str, optional
        An optional list of file names to pass directly to the manual loader in
        "show_list" mode. When provided, the function will include the filenames
        in the options string instead of relying on a file pattern; items should
        be either full paths or relative to the selected `source_directory`.
    """
    # xml_filename = project_filename + ".xml"

    if definition_opts is None:
        definition_opts = DefinitionOptions()

    show_list_options = "" if not list_files else "show_list " + " ".join(list_files)

    options = (
        "define_dataset=[Manual Loader (Bioformats based)] "
        + "project_filename=["
        + project_filename
        + "] "
        + "_____"
        + definition_opts.fmt_acitt_options()
        + " "
        + "image_file_directory="
        + source_directory
        + " "
        + "image_file_pattern="
        + image_file_pattern
        + " "
        + dataset_organisation
        + " "
        + "calibration_type=[Same voxel-size for all views] "
        + "calibration_definition=[Load voxel-size(s) from file(s)] "
        + show_list_options
    )

    log.debug("Manual dataset definition options: <%s>", options)
    IJ.run("Define Multi-View Dataset", str(options))


def resave_as_h5(
    source_xml_file,
    output_h5_file_path,
    processing_opts=None,
    timepoints_per_partition=1,
    use_deflate_compression=True,
    subsampling_factors=None,
    hdf5_chunk_sizes=None,
):
    """Resave the xml dataset in a new format (either all or single timepoints).

    Useful if it hasn't been done during dataset definition (see
    `define_dataset_auto()`). Allows e.g. parallelization of HDF-5 re-saving.

    Parameters
    ----------
    source_xml_file : File or str
        XML input file.
    output_h5_file_path : str
        Export path for the output file including the `.xml `extension.
    processing_opts : imcflibs.imagej.bdv.ProcessingOptions, optional
        The `ProcessingOptions` object defining parameters for the run. Will
        fall back to the defaults defined in the corresponding class if the
        parameter is `None` or skipped.
    timepoints_per_partition : int, optional
        How many timepoints to export per partition, by default `1`.
    use_deflate_compression : bool, optional
        Run deflate compression, by default `True`.
    subsampling_factors : str, optional
        Specify subsampling factors explicitly, for example:
        `[{ {1,1,1}, {2,2,1}, {4,4,2}, {8,8,4} }]`.
    hdf5_chunk_sizes : str, optional
        Specify hdf5_chunk_sizes factors explicitly, for example
        `[{ {32,16,8}, {16,16,16}, {16,16,16}, {16,16,16} }]`.
    """

    if not processing_opts:
        processing_opts = ProcessingOptions()

    if use_deflate_compression:
        use_deflate_compression_arg = "use_deflate_compression "
    else:
        use_deflate_compression_arg = ""

    # If split_hdf5 option
    if timepoints_per_partition != 0:
        split_hdf5 = "split_hdf5 "
    else:
        split_hdf5 = ""

    if subsampling_factors:
        subsampling_factors = "subsampling_factors=" + subsampling_factors + " "
    else:
        subsampling_factors = " "
    if hdf5_chunk_sizes:
        hdf5_chunk_sizes = "hdf5_chunk_sizes=" + hdf5_chunk_sizes + " "
    else:
        hdf5_chunk_sizes = " "

    options = (
        "select="
        + str(source_xml_file)
        + " "
        + processing_opts.fmt_acitt_options("resave")
        + processing_opts.fmt_acitt_selectors()
        + subsampling_factors
        + hdf5_chunk_sizes
        + "timepoints_per_partition="
        + str(timepoints_per_partition)
        + " "
        + "setups_per_partition=0 "
        + use_deflate_compression_arg
        + split_hdf5
        + "export_path="
        + output_h5_file_path
    )

    log.debug("Resave as HDF5 options: <%s>", options)
    IJ.run("Resave as HDF5 (local)", str(options))


def flip_axes(source_xml_file, x=False, y=True, z=False):
    """Call BigStitcher's "Flip Axes" command.

    Wrapper for `BigStitcher > Batch Processing > Tools > Flip Axes`. This is
    required for some formats, for example Nikon `.nd2` files need a flip along
    the Y-axis.

    Parameters
    ----------
    source_xml_file : str
        Full path to the `.xml` file.
    x : bool, optional
        Flip images along the X-axis, by default `False`.
    y : bool, optional
        Flip mages along the Y-axis, by default `True`.
    z : bool, optional
        Flip images along the Z-axis, by default `False`.
    """

    file_info = pathtools.parse_path(source_xml_file)

    axes_to_flip = ""
    if x is True:
        axes_to_flip += " flip_x"
    if y is True:
        axes_to_flip += " flip_y"
    if z is True:
        axes_to_flip += " flip_z"

    IJ.run("Flip Axes", "select=" + source_xml_file + axes_to_flip)

    backup_xml_files(file_info["path"], "flip_axes")


def phase_correlation_pairwise_shifts_calculation(
    project_path,
    processing_opts=None,
    downsampling_xyz="",
):
    """Calculate pairwise shifts using Phase Correlation.

    Parameters
    ----------
    project_path : str
        Full path to the `.xml` file.
    processing_opts : imcflibs.imagej.bdv.ProcessingOptions, optional
        The `ProcessingOptions` object defining parameters for the run. Will
        fall back to the defaults defined in the corresponding class if the
        parameter is `None` or skipped.
    downsampling_xyz : list of int, optional
        Downsampling factors in X, Y and Z, for example `[4,4,4]`. By default
        empty which will result in BigStitcher choosing the factors.
    """

    if not processing_opts:
        processing_opts = ProcessingOptions()

    file_info = pathtools.parse_path(project_path)

    if downsampling_xyz != "":
        downsampling = "downsample_in_x=%s downsample_in_y=%s downsample_in_z=%s " % (
            downsampling_xyz[0],
            downsampling_xyz[1],
            downsampling_xyz[2],
        )
    else:
        downsampling = ""

    options = (
        "select=["
        + project_path
        + "] "
        + processing_opts.fmt_acitt_options()
        + processing_opts.fmt_acitt_selectors()
        + " "
        + "method=[Phase Correlation] "
        + "show_expert_grouping_options "
        + "show_expert_algorithm_parameters "
        + processing_opts.fmt_use_acitt()
        + processing_opts.fmt_how_to_treat()
        + downsampling
        + "subpixel_accuracy"
    )

    log.debug("Calculate pairwise shifts options: <%s>", options)
    IJ.run("Calculate pairwise shifts ...", str(options))

    backup_xml_files(file_info["path"], "phase_correlation_shift_calculation")


def filter_pairwise_shifts(
    project_path,
    min_r=0.7,
    max_r=1,
    max_shift_xyz="",
    max_displacement="",
):
    """Filter the pairwise shifts based on different thresholds.

    Parameters
    ----------
    project_path : str
        Path to the `.xml` on which to apply the filters.
    min_r : float, optional
        Minimal quality of the link to keep, by default `0.7`.
    max_r : float, optional
        Maximal quality of the link to keep, by default `1`.
    max_shift_xyz : list(int), optional
        Maximal shift in X, Y and Z (in pixels) to keep, e.g. `[10,10,10]`. By
        default empty, meaning no filtering based on the shifts will be applied.
    max_displacement : int, optional
        Maximal displacement to keep. By default empty, meaning no filtering
        based on the displacement will be applied.
    """

    file_info = pathtools.parse_path(project_path)

    if max_shift_xyz != "":
        filter_by_max_shift = (
            " filter_by_shift_in_each_dimension"
            " max_shift_in_x=%s max_shift_in_y=%s max_shift_in_z=%s"
        ) % (max_shift_xyz[0], max_shift_xyz[1], max_shift_xyz[2])
    else:
        filter_by_max_shift = ""

    if max_displacement != "":
        filter_by_max_displacement = (
            " filter_by_total_shift_magnitude max_displacement=%s"
        ) % (max_displacement)
    else:
        filter_by_max_displacement = ""

    options = (
        "select=["
        + project_path
        + "] "
        + "filter_by_link_quality "
        + "min_r="
        + str(min_r)
        + " "
        + "max_r="
        + str(max_r)
        + filter_by_max_shift
        + filter_by_max_displacement
    )

    log.debug("Filter pairwise options: <%s>", options)
    IJ.run("Filter pairwise shifts ...", str(options))

    backup_xml_files(file_info["path"], "filter_pairwise_shifts")


def optimize_and_apply_shifts(
    project_path,
    processing_opts=None,
    relative_error=2.5,
    absolute_error=3.5,
):
    """Optimize the shifts and apply them to the dataset.

    Parameters
    ----------
    project_path : str
        Path to the `.xml` on which to optimize and apply the shifts.
    processing_opts : imcflibs.imagej.bdv.ProcessingOptions, optional
        The `ProcessingOptions` object defining parameters for the run. Will
        fall back to the defaults defined in the corresponding class if the
        parameter is `None` or skipped.
    relative_error : float, optional
        Relative alignment error (in px) to accept, by default `2.5`.
    absolute_error : float, optional
        Absolute alignment error (in px) to accept, by default `3.5`.
    """

    if not processing_opts:
        processing_opts = ProcessingOptions()

    file_info = pathtools.parse_path(project_path)

    options = (
        "select=["
        + project_path
        + "] "
        + processing_opts.fmt_acitt_options()
        + processing_opts.fmt_acitt_selectors()
        + " "
        + "relative="
        + str(relative_error)
        + " "
        + "absolute="
        + str(absolute_error)
        + " "
        + "global_optimization_strategy=[Two-Round using Metadata to align unconnected "
        + "Tiles and iterative dropping of bad links] "
        + "show_expert_grouping_options "
        + processing_opts.fmt_use_acitt()
        + processing_opts.fmt_how_to_treat()
    )

    log.debug("Optimization and shifts application options: <%s>", options)
    IJ.run("Optimize globally and apply shifts ...", str(options))

    backup_xml_files(file_info["path"], "optimize_and_apply_shifts")


def detect_interest_points(
    project_path,
    processing_opts=None,
    sigma=1.8,
    threshold=0.008,
    maximum_number=3000,
):
    """Run the "Detect Interest Points" command for registration.

    Parameters
    ----------
    project_path : str
        Path to the `.xml` project.
    processing_opts : imcflibs.imagej.bdv.ProcessingOptions, optional
        The `ProcessingOptions` object defining parameters for the run. Will
        fall back to the defaults defined in the corresponding class if the
        parameter is `None` or skipped.
    sigma : float, optional
        Minimum sigma for interest points detection, by default `1.8`.
    threshold : float, optional
        Threshold value for the interest point detection, by default `0.008`.
    maximum_number : int, optional
        Maximum number of interest points to use, by default `3000`.
    """

    if not processing_opts:
        processing_opts = ProcessingOptions()

    options = (
        "select=["
        + project_path
        + "] "
        + processing_opts.fmt_acitt_options()
        + processing_opts.fmt_acitt_selectors()
        + "type_of_interest_point_detection=Difference-of-Gaussian "
        + "label_interest_points=beads "
        + "limit_amount_of_detections "
        + "group_tiles "
        + "subpixel_localization=[3-dimensional quadratic fit] "
        + "interest_point_specification=[Advanced ...] "
        + "downsample_xy=8x "
        + "downsample_z=2x "
        + "sigma="
        + str(sigma)
        + " "
        + "threshold="
        + str(threshold)
        + " "
        + "find_maxima "
        + "maximum_number="
        + str(maximum_number)
        + " "
        + "type_of_detections_to_use=Brightest "
        + "compute_on=[CPU (Java)]"
    )

    log.debug("Interest points detection options: <%s>", options)
    IJ.run("Detect Interest Points for Registration", str(options))


def interest_points_registration(
    project_path,
    processing_opts=None,
    rigid_timepoints=False,
):
    """Run the "Register Dataset based on Interest Points" command.

    Parameters
    ----------
    project_path : str
        Path to the `.xml` project.
    processing_opts : imcflibs.imagej.bdv.ProcessingOptions, optional
        The `ProcessingOptions` object defining parameters for the run. Will
        fall back to the defaults defined in the corresponding class if the
        parameter is `None` or skipped. This controls which angles, channels,
        illuminations, tiles and timepoints are processed.
    rigid_timepoints : bool, optional
        If set to `True` each timepoint will be considered as a rigid unit
        (useful e.g. if spatial registration has already been performed before).
        By default `False`.
    """

    if not processing_opts:
        processing_opts = ProcessingOptions()

    if rigid_timepoints:
        rigid_timepoints_arg = "consider_each_timepoint_as_rigid_unit "
    else:
        rigid_timepoints_arg = " "

    options = (
        "select=["
        + project_path
        + "] "
        + processing_opts.fmt_acitt_options()
        + processing_opts.fmt_acitt_selectors()
        + "registration_algorithm=[Precise descriptor-based (translation invariant)] "
        + "registration_over_time=["
        + "Match against one reference timepoint (no global optimization)] "
        + "registration_in_between_views=["
        + "Only compare overlapping views (according to current transformations)] "
        + "interest_point_inclusion=[Compare all interest point of overlapping views] "
        + "interest_points=beads "
        + "group_tiles "
        + "group_illuminations "
        + "group_channels "
        + "reference=1 "
        + rigid_timepoints_arg
        + "transformation=Affine "
        + "regularize_model "
        + "model_to_regularize_with=Affine "
        + "lambda=0.10 "
        + "number_of_neighbors=3 "
        + "redundancy=1 "
        + "significance=3 "
        + "search_radius=100 "
        + "allowed_error_for_ransac=5 "
        + "inlier_factor=3 "
        + "ransac_iterations=Normal "
        + "global_optimization_strategy=["
        + "Two-Round: Handle unconnected tiles, "
        + "remove wrong links RELAXED (5.0x / 7.0px)] "
        + "interestpoint_grouping=["
        + "Group interest points (simply combine all in one virtual view)] "
        + "interest=5"
    )

    log.debug("Interest points registration options: <%s>", options)
    # register using interest points
    IJ.run("Register Dataset based on Interest Points", options)


def duplicate_transformations(
    project_path,
    transformation_type="channel",
    channel_source=0,
    tile_source=None,
    transformation_to_use="[Replace all transformations]",
):
    """Duplicate / propagate transformation parameters to other channels.

    Propagate the transformation parameters generated by a previously performed
    registration of a single channel to the other channels.

    Parameters
    ----------
    project_path : str
        Path to the `.xml` project.
    transformation_type : str, optional
        Transformation mode, one of `channel` (to propagate from one channel to
        all others) and `tiles` (to propagate from one tile to all others).
    channel_source : int, optional
        Reference channel number (starting at 1), by default None.
    tile_source : int, optional
        Reference tile, by default None.
    transformation_to_use : str, optional
        One of `[Replace all transformations]` (default) and `[Add last
        transformation only]` to specify which transformations to propagate.
    """

    file_info = pathtools.parse_path(project_path)

    apply = ""
    source = ""
    target = ""
    tile_apply = ""
    tile_process = ""

    ch_apply = ""
    ch_process = ""

    if transformation_type == "channel":
        apply = "[One channel to other channels]"
        target = "[All Channels]"
        if channel_source > 0:
            source = str(channel_source - 1)
        else:
            source = "0"
        if tile_source:
            tile_apply = "apply_to_tile=[Single tile (Select from List)] "
            tile_process = "processing_tile=[tile " + str(tile_source) + "] "
        else:
            tile_apply = "apply_to_tile=[All tiles] "
    elif transformation_type == "tile":
        apply = "[One tile to other tiles]"
        target = "[All Tiles]"
        source = str(tile_source)
        if channel_source > 0:
            ch_apply = "apply_to_channel=[Single channel (Select from List)] "
            ch_process = "processing_channel=[channel " + str(channel_source - 1) + "] "
        else:
            ch_apply = "apply_to_channel=[All channels] "
    else:
        raise ValueError("Invalid transformation type: %s" % transformation_type)

    options = (
        "apply="
        + apply
        + " "
        + "select=["
        + project_path
        + "] "
        + "apply_to_angle=[All angles] "
        + "apply_to_illumination=[All illuminations] "
        + tile_apply
        + tile_process
        + ch_apply
        + ch_process
        + "apply_to_timepoint=[All Timepoints] "
        + "source="
        + source
        + " "
        + "target="
        + target
        + " "
        + "duplicate_which_transformations="
        + transformation_to_use
        + " "
    )

    log.debug("Transformation duplication options: <%s>", options)
    IJ.run("Duplicate Transformations", str(options))

    backup_xml_files(
        file_info["path"],
        "duplicate_transformation_" + transformation_type,
    )


def fuse_dataset(
    project_path,
    processing_opts=None,
    result_path=None,
    downsampling=1,
    interpolation="[Linear Interpolation]",
    pixel_type="[16-bit unsigned integer]",
    fusion_type="Avg, Blending",
    export="HDF5",
    compression="Zstandard",
):
    """Call BigStitcher's "Fuse Dataset" command.

    Wrapper to `BigStitcher > Batch Processing > Fuse Dataset`.

    Depending on the export type, inputs are different and therefore will
    distribute inputs differently.

    Parameters
    ----------
    project_path : str
        Path to the `.xml` on which to run the fusion.
    processing_opts : imcflibs.imagej.bdv.ProcessingOptions, optional
        The `ProcessingOptions` object defining parameters for the run. Will
        fall back to the defaults defined in the corresponding class if the
        parameter is `None` or skipped.
    result_path : str, optional
        Path to store the resulting fused image, by default `None` which will
        store the result in the same folder as the input project.
    downsampling : int, optional
        Downsampling value to use during fusion, by default `1`.
    interpolation : str, optional
        Interpolation to use during fusion, by default `[Linear Interpolation]`.
    pixel_type : str, optional
        Pixel type to use during fusion, by default `[16-bit unsigned integer]`.
    export : str, optional
        Format of the output fused image, by default `HDF5`.
    fusion_type : str, optional
        Type of fusion algorithm to use, by default `Avg, Blending`.
    compression : str, optional
        Compression method to use when exporting as HDF5, by default `Zstandard`.
    """

    if processing_opts is None:
        processing_opts = ProcessingOptions()

    file_info = pathtools.parse_path(project_path)
    if not result_path:
        result_path = file_info["path"]
        # if not os.path.exists(result_path):
        #     os.makedirs(result_path)

    options = (
        "select=["
        + project_path
        + "] "
        + processing_opts.fmt_acitt_options()
        + "bounding_box=[All Views] "
        + "downsampling="
        + str(downsampling)
        + " "
        + "interpolation="
        + interpolation
        + " "
        + "fusion_type=["
        + fusion_type
        + "] "
        + "pixel_type="
        + pixel_type
        + " "
        + "interest_points_for_non_rigid=[-= Disable Non-Rigid =-] "
        + "preserve_original "
        + "produce=[Each timepoint & channel] "
    )

    if export == "TIFF":
        options = (
            options
            + "fused_image=[Save as (compressed) TIFF stacks] "
            + "define_input=[Auto-load from input data (values shown below)] "
            + "output_file_directory=["
            + result_path
            + "/.] "
            + "filename_addition=["
            + file_info["basename"]
            + "]"
        )
    elif export == "HDF5":
        h5_fused_path = pathtools.join2(
            result_path, file_info["basename"] + "_fused.h5"
        )
        xml_fused_path = pathtools.join2(
            result_path, file_info["basename"] + "_fused.xml"
        )

        options = (
            options
            + "fused_image=[OME-ZARR/N5/HDF5 export using N5-API] "
            + "define_input=[Auto-load from input data (values shown below)] "
            + "export=HDF5 "
            + "compression="
            + compression
            + " "
            + "create "
            + "create_0 "
            + "hdf5_file=["
            + h5_fused_path
            + "] "
            + "xml_output_file=["
            + xml_fused_path
            + "] "
            + "show_advanced_block_size_options "
            + "block_size_x=128 "
            + "block_size_y=128 "
            + "block_size_z=64 "
            + "block_size_factor_x=1 "
            + "block_size_factor_y=1 "
            + "block_size_factor_z=1"
        )

    log.debug("Dataset fusion options: <%s>", options)
    IJ.run("Image Fusion", str(options))


def fuse_dataset_bdvp(
    project_path,
    command,
    result_path=None,
    fusion_method="SMOOTH AVERAGE",
    range_channels="",
    range_slices="",
    range_frames="",
    n_resolution_levels=5,
    use_lzw_compression=True,
    split_slices=False,
    split_channels=False,
    split_frames=False,
    override_z_ratio=False,
    z_ratio=1.0,
    use_interpolation=True,
):
    """Export a project using the BigDataViewer playground (`bdvp`) exporter.

    Use the BigDataViewer playground / BIOP Kheops exporter to fuse a
    BigDataViewer project and save it as pyramidal OME-TIFF.

    Parameters
    ----------
    project_path : str
        Full path to the BigDataViewer XML project file.
    command : CommandService
        The Scijava CommandService instance to execute the export command.
    result_path : str, optional
        Path where to store the exported files. If `None`, files will be
        saved in the same directory as the input project.
    fusion_method : str, optional
        Fusion method to use for exporting (default `SMOOTH AVERAGE`).
    range_channels : str, optional
        Channels to include in the export. Default is all channels.
    range_slices : str, optional
        Slices to include in the export. Default is all slices.
    range_frames : str, optional
        Frames to include in the export. Default is all frames.
    n_resolution_levels : int, optional
        Number of pyramid resolution levels to use for the export. Default is 5.
    use_lzw_compression : bool, optional
        Compress the output file using LZW. Default is True.
    split_slices : bool, optional
        Split output into separate files for each slice. Default is False.
    split_channels : bool, optional
        Split output into separate files for each channel. Default is False.
    split_frames : bool, optional
        Split output into separate files for each frame. Default is False.
    override_z_ratio : bool, optional
        Override the default `z_ratio` value. Default is False.
    z_ratio : float, optional
        The z ratio to use for the export. Default is 1.0.
    use_interpolation : bool, optional
        Interpolate during fusion (takes ~4x longer). Default is True.

    Notes
    -----
    Requires the `PTBIOP` update site to be enabled in Fiji/ImageJ.

    Examples
    --------
    Example 1 - simple export using a CommandService instance available as
    `command`, using the default options and placing the output  next to the
    input xml:

    >>> #@ CommandService command
    >>> xml_input = "/path/to/project.xml"
    >>> fuse_dataset_bdvp(xml_input, command)

    Example 2 - explicit options using a custom output path, specific channels,
    disabling interpolation and overriding the z-ratio:

    >>> #@ CommandService command
    >>> xml_input = "/path/to/project.xml"
    >>> out_dir = "/path/to/output_dir"
    >>> fuse_dataset_bdvp(
    ...     xml_input,
    ...     command,
    ...     result_path=out_dir,
    ...     fusion_method="SMOOTH AVERAGE",
    ...     range_channels="0-1",
    ...     n_resolution_levels=4,
    ...     use_lzw_compression=False,
    ...     split_channels=True,
    ...     override_z_ratio=True,
    ...     z_ratio=2.0,
    ...     use_interpolation=False,
    ... )
    """
    file_info = pathtools.parse_path(project_path)

    if not result_path:
        result_path = file_info["path"]

    command.run(
        FuseBigStitcherDatasetIntoOMETiffCommand,
        True,  # seems to indicate whether to run the command headless or not
        "xml_bigstitcher_file",
        project_path,
        "output_path_directory",
        result_path,
        "range_channels",
        range_channels,
        "range_slices",
        range_slices,
        "range_frames",
        range_frames,
        "n_resolution_levels",
        n_resolution_levels,
        "fusion_method",
        fusion_method,
        "use_lzw_compression",
        use_lzw_compression,
        "split_slices",
        split_slices,
        "split_channels",
        split_channels,
        "split_frames",
        split_frames,
        "override_z_ratio",
        override_z_ratio,
        "z_ratio",
        z_ratio,
        "use_interpolation",
        use_interpolation,
    ).get()



"""Functions to work with results tables."""

from ij.measure import ResultsTable


def preset_results_column(results_table, column, value):
    """Pre-set all rows in given column of the ResultsTable with desired values.

    Parameters
    ----------
    results_table : ij.measure.ResultsTable
        a reference of the IJ-ResultsTable
    column : str
        the desired column. will be created if it does not yet exist
    value : str, float or int
        the value to be set
    """
    for i in range(results_table.size()):
        results_table.setValue(column, i, value)

    results_table.show("Results")


def add_results_to_resultstable(results_table, column, values, rows=None):
    """Add values to the ResultsTable in a specified column.

    This function works in two ways, depending on the value of `rows`:

    1. If rows is `None`, it adds values sequentially starting from row 0.
    2. If rows is a list of int, it adds values to the given row indices.

    Parameters
    ----------
    results_table : ij.measure.ResultsTable
        A reference to the IJ-ResultsTable
    column : str
        The column in which to add the values.
    values : list of int, float or str
        Values to be added.
    rows : list of int, optional
        Specific row indices where values should be added. If None, values are
        added sequentially starting from row 0.

    Examples
    --------
    To add the same value (42) to a given `ResultsTable` to rows 1, 3 and 5:
    >>> add_results_to_resultstable(rt, "Intensity", 42, rows=[1, 3, 5])
    """
    if not isinstance(values, list) and rows is not None:
        values = [values] * len(rows)

    # Case 1: Add values sequentially from row 0
    if rows is None:
        for index, value in enumerate(values):
            results_table.setValue(column, index, value)

    # Case 2: Add values to specific rows
    else:
        if len(values) != len(rows):
            raise ValueError(f"Length mismatch: values ({len(values)}) and rows ({len(rows)})")

        for i, row_index in enumerate(rows):
            results_table.setValue(column, row_index, values[i])

    results_table.show("Results")



def get_resultstable():
    """Instantiate or get the ResultsTable instance.

    Use to either get the current instance of the IJ ResultsTable or instantiate
    it if it does not yet exist.

    Returns
    -------
    ij.measure.ResultsTable
        A reference of the IJ-ResultsTable
    """
    rt = ResultsTable.getInstance()
    if not rt:
        rt = ResultsTable()
    return rt



"""Functions working with [TrackMate].

[TrackMate]: https://imagej.net/plugins/trackmate/
"""

import os
import sys

from fiji.plugin.trackmate import Logger, Model, SelectionModel, Settings, TrackMate
from fiji.plugin.trackmate.action import LabelImgExporter
from fiji.plugin.trackmate.action.LabelImgExporter import LabelIdPainting
from fiji.plugin.trackmate.cellpose import CellposeDetectorFactory
from fiji.plugin.trackmate.cellpose.CellposeSettings import PretrainedModel
from fiji.plugin.trackmate.detection import LogDetectorFactory
from fiji.plugin.trackmate.features import FeatureFilter
from fiji.plugin.trackmate.stardist import StarDistDetectorFactory
from fiji.plugin.trackmate.tracking.jaqaman import SparseLAPTrackerFactory
from ij import IJ
from java.lang import Double

from .. import pathtools


def cellpose_detector(
    imageplus,
    cellpose_env_path,
    model_to_use,
    obj_diameter,
    target_channel,
    optional_channel=0,
    use_gpu=True,
    simplify_contours=True,
):
    """Create a dictionary with all settings for TrackMate using Cellpose.

    Parameters
    ----------
    imageplus : ij.ImagePlus
        ImagePlus on which to apply the detector.
    cellpose_env_path : str
        Path to the Cellpose environment.
    model_to_use : str
        Name of the model to use for the segmentation (CYTO, NUCLEI, CYTO2).
    obj_diameter : float
        Diameter of the objects to detect in the image.
        This will be calibrated to the unit used in the image.
    target_channel : int
        Index of the channel to use for segmentation.
    optional_channel : int, optional
        Index of the secondary channel to use for segmentation, by default 0.
    use_gpu : bool, optional
        Boolean for GPU usage, by default True.
    simplify_contours : bool, optional
        Boolean for simplifying the contours, by default True.

    Returns
    -------
    fiji.plugin.trackmate.Settings
        Dictionary containing all the settings to use for TrackMate.

    Example
    -------
    >>> settings = cellpose_detector(
    ...    imageplus=imp,
    ...    cellpose_env_path="D:/CondaEnvs/cellpose",
    ...    model_to_use="NUCLEI",
    ...    obj_diameter=23.0,
    ...    target_channel=1,
    ...    optional_channel=0
    ... )
    """
    settings = Settings(imageplus)

    settings.detectorFactory = CellposeDetectorFactory()
    settings.detectorSettings["TARGET_CHANNEL"] = target_channel
    # set optional channel to 0, will be overwritten if needed:
    settings.detectorSettings["OPTIONAL_CHANNEL_2"] = optional_channel

    settings.detectorSettings["CELLPOSE_PYTHON_FILEPATH"] = pathtools.join2(
        cellpose_env_path, "python.exe"
    )
    settings.detectorSettings["CELLPOSE_MODEL_FILEPATH"] = os.path.join(
        os.environ["USERPROFILE"], ".cellpose", "models"
    )
    input_to_model = {
        "nuclei": PretrainedModel.NUCLEI,
        "cyto": PretrainedModel.CYTO,
        "cyto2": PretrainedModel.CYTO2,
    }
    if model_to_use.lower() in input_to_model:
        selected_model = input_to_model[model_to_use.lower()]
    else:
        print("Selected Model Does Not Exist")
        return

    settings.detectorSettings["CELLPOSE_MODEL"] = selected_model
    settings.detectorSettings["CELL_DIAMETER"] = obj_diameter
    settings.detectorSettings["USE_GPU"] = use_gpu
    settings.detectorSettings["SIMPLIFY_CONTOURS"] = simplify_contours

    return settings


def stardist_detector(imageplus, target_chnl):
    """Create a dictionary with all settings for TrackMate using StarDist.

    Parameters
    ----------
    imageplus : ij.ImagePlus
        Image on which to do the segmentation.
    target_chnl : int
        Index of the channel on which to do the segmentation.

    Returns
    -------
    fiji.plugin.trackmate.Settings
        Dictionary containing all the settings to use for TrackMate.
    """

    settings = Settings(imageplus)
    settings.detectorFactory = StarDistDetectorFactory()
    settings.detectorSettings["TARGET_CHANNEL"] = target_chnl

    return settings


def log_detector(
    imageplus,
    radius,
    target_channel,
    quality_threshold=0.0,
    median_filtering=True,
    subpix_localization=True,
):
    """Create a dictionary with all settings for TrackMate using the LogDetector.

    Parameters
    ----------
    imageplus : ij.ImagePlus
        Image on which to do the segmentation.
    radius : float
        Radius of the objects to detect.
    target_channel : int
        Index of the channel on which to do the segmentation.
    quality_threshold : int, optional
        Threshold to use for excluding the spots by quality, by default 0.
    median_filtering : bool, optional
        Boolean to do median filtering, by default True.
    subpix_localization : bool, optional
        Boolean to do subpixel localization, by default True.

    Returns
    -------
    fiji.plugin.trackmate.Settings
        Dictionary containing all the settings to use for TrackMate.
    """

    settings = Settings(imageplus)
    settings.detectorFactory = LogDetectorFactory()

    settings.detectorSettings["RADIUS"] = Double(radius)
    settings.detectorSettings["TARGET_CHANNEL"] = target_channel
    settings.detectorSettings["THRESHOLD"] = Double(quality_threshold)
    settings.detectorSettings["DO_MEDIAN_FILTERING"] = median_filtering
    settings.detectorSettings["DO_SUBPIXEL_LOCALIZATION"] = subpix_localization

    return settings


def spot_filtering(
    settings,
    quality_thresh=None,
    area_thresh=None,
    circularity_thresh=None,
    intensity_dict_thresh=None,
):
    """Add spot filtering for different features to the settings dictionary.

    Parameters
    ----------
    settings : fiji.plugin.trackmate.Settings
        Dictionary containing all the settings to use for TrackMate.
    quality_thresh : float, optional
        Threshold to use for quality filtering of the spots, by default None.
        If the threshold is positive, will exclude everything below the value.
        If the threshold is negative, will exclude everything above the value.
    area_thresh : float, optional
        Threshold to use for area filtering of the spots, keep None with LoG Detector -
        by default also None.
        If the threshold is positive, will exclude everything below the value.
        If the threshold is negative, will exclude everything above the value.
    circularity_thresh : float, optional
        Threshold to use for circularity thresholding (needs to be between 0 and 1, keep None with LoG Detector)
        - by default None.
        If the threshold is positive, will exclude everything below the value.
        If the threshold is negative, will exclude everything above the value.
    intensity_dict_thresh : dict, optional
        Threshold to use for intensity filtering of the spots, by default None.
        Dictionary needs to contain the channel index as key and the filter as value.
        If the threshold is positive, will exclude everything below the value.
        If the threshold is negative, will exclude everything above the value.

    Returns
    -------
    fiji.plugin.trackmate.Settings
        Dictionary containing all the settings to use for TrackMate.
    """

    settings.initialSpotFilterValue = -1.0
    settings.addAllAnalyzers()

    # Here 'true' takes everything ABOVE the mean_int value
    if quality_thresh:
        filter_spot = FeatureFilter(
            "QUALITY",
            Double(abs(quality_thresh)),
            quality_thresh >= 0,
        )
        settings.addSpotFilter(filter_spot)
    if area_thresh:  # Keep none for log detector
        filter_spot = FeatureFilter("AREA", Double(abs(area_thresh)), area_thresh >= 0)
        settings.addSpotFilter(filter_spot)
    if circularity_thresh:  # has to be between 0 and 1, keep none for log detector
        filter_spot = FeatureFilter(
            "CIRCULARITY", Double(abs(circularity_thresh)), circularity_thresh >= 0
        )
        settings.addSpotFilter(filter_spot)
    if intensity_dict_thresh:
        for key, value in intensity_dict_thresh.items():
            filter_spot = FeatureFilter(
                "MEAN_INTENSITY_CH" + str(key), abs(value), value >= 0
            )
            settings.addSpotFilter(filter_spot)

    return settings


def set_spotfilter(settings, filter_key, filter_value):
    """Set a TrackMate spot filter with specified filter key and values.

    Parameters
    ----------
    settings : fiji.plugin.trackmate.Settings
        Settings object to use for TrackMate.
    filter_key : str
        The key-name of the filter to be applied (as opposed to the filter
        "name" shown in ImageJ. Refer to the spot features table on the related
        ImageJ wiki page:
        https://imagej.net/plugins/trackmate/scripting/trackmate-detectors-trackers-keys#the-feature-penalty-map
    filter_value : list
        A list containing two values for the filter. The first value is applied
        as an above-threshold filter, and the second as a below-threshold
        filter.

    Returns
    -------
    Settings
        The modified TrackMate settings dict with added spot filters

    Example
    -------

    To set an above-threshold filter value for spot `QUALITY` without a
    below-threshold value use:

    >>> settings = set_trackmate_spotfilter(tm_settings, 'QUALITY', [120, None])
    """
    settings.addAllAnalyzers()
    if filter_value[0] != None:
        filter_low = FeatureFilter(filter_key, filter_value[0], True)
        settings.addSpotFilter(filter_low)
    if filter_value[1] != None:
        filter_high = FeatureFilter(filter_key, filter_value[1], False)
        settings.addSpotFilter(filter_high)

    return settings


def sparse_lap_tracker(settings):
    """Create a sparse LAP tracker with default settings.

    Parameters
    ----------
    settings : fiji.plugin.trackmate.Settings
        Dictionary containing all the settings to use for TrackMate.

    Returns
    -------
    fiji.plugin.trackmate.Settings
        Dictionary containing all the settings to use for TrackMate.
    """

    settings.trackerFactory = SparseLAPTrackerFactory()
    settings.trackerSettings = settings.trackerFactory.getDefaultSettings()

    return settings


def track_filtering(
    settings,
    link_max_dist=15.0,
    gap_closing_dist=15.0,
    max_frame_gap=3,
    track_splitting_max_dist=None,
    track_merging_max_distance=None,
):
    """Add track filtering for different features to the settings dictionary.

    Parameters
    ----------
    settings : fiji.plugin.trackmate.Settings
        Dictionary containing all the settings to use for TrackMate.
    link_max_dist : float, optional
        Maximal displacement of the spots, by default 0.5.
    gap_closing_dist : float, optional
        Maximal distance for gap closing, by default 0.5.
    max_frame_gap : int, optional
        Maximal frame interval between spots to be bridged, by default 2.
    track_splitting_max_dist : int, optional
        Maximal frame interval for splitting tracks, by default None.
    track_merging_max_distance : int, optional
        Maximal frame interval for merging tracks , by default None.

    Returns
    -------
    fiji.plugin.trackmate.Settings
        Dictionary containing all the settings to use for TrackMate.
    """
    # NOTE: `link_max_dist` and `gap_closing_dist` must be double!
    settings.trackerSettings["LINKING_MAX_DISTANCE"] = link_max_dist
    settings.trackerSettings["GAP_CLOSING_MAX_DISTANCE"] = gap_closing_dist
    settings.trackerSettings["MAX_FRAME_GAP"] = max_frame_gap
    if track_splitting_max_dist:
        settings.trackerSettings["ALLOW_TRACK_SPLITTING"] = True
        settings.trackerSettings["SPLITTING_MAX_DISTANCE"] = track_splitting_max_dist
    if track_merging_max_distance:
        settings.trackerSettings["ALLOW_TRACK_MERGING"] = True
        settings.trackerSettings["MERGING_MAX_DISTANCE"] = track_merging_max_distance

    return settings


def run_trackmate(
    implus,
    settings,
    crop_roi=None,
):
    # sourcery skip: merge-else-if-into-elif, swap-if-else-branches
    """Run TrackMate on an open ImagePlus object.

    Parameters
    ----------
    implus : ij.ImagePlus
        ImagePlus image on which to run Trackmate.
    settings : fiji.plugin.trackmate.Settings
        Settings to use for TrackMate, see detector methods for different settings.
    crop_roi : ij.gui.Roi, optional
        ROI to crop on the image, by default None.

    Returns
    -------
    ij.ImagePlus
        Labeled image with all the objects belonging to the same tracks having
        the same label.
    """

    dims = implus.getDimensions()
    cal = implus.getCalibration()

    if implus.getNSlices() > 1:
        implus.setDimensions(dims[2], dims[4], dims[3])

    if crop_roi is not None:
        implus.setRoi(crop_roi)

    model = Model()

    model.setLogger(Logger.IJTOOLBAR_LOGGER)

    # Configure tracker
    # settings.addTrackAnalyzer(TrackDurationAnalyzer())
    settings.initialSpotFilterValue = -1.0

    trackmate = TrackMate(model, settings)
    trackmate.computeSpotFeatures(True)
    trackmate.computeTrackFeatures(True)

    if not settings.trackerFactory:
        # Create a Sparse LAP Tracker if no Tracker has been created
        settings = sparse_lap_tracker(settings)

    ok = trackmate.checkInput()
    if not ok:
        sys.exit(str(trackmate.getErrorMessage()))

    ok = trackmate.process()
    if not ok:
        if "[SparseLAPTracker] The spot collection is empty." in str(
            trackmate.getErrorMessage()
        ):
            new_imp = IJ.createImage(
                "Untitled",
                str(implus.getBitDepth()) + "-bit black",
                implus.getWidth(),
                implus.getHeight(),
                implus.getNFrames(),
            )
            new_imp.setCalibration(cal)

            return new_imp

        else:
            sys.exit(str(trackmate.getErrorMessage()))

    SelectionModel(model)

    exportSpotsAsDots = False
    exportTracksOnly = False
    labelIdPainting = LabelIdPainting.LABEL_IS_TRACK_ID
    # implus2.close()
    label_imp = LabelImgExporter.createLabelImagePlus(
        trackmate, exportSpotsAsDots, exportTracksOnly, labelIdPainting
    )
    label_imp.setCalibration(cal)
    label_imp.setDimensions(dims[2], dims[3], dims[4])
    implus.setDimensions(dims[2], dims[3], dims[4])

    return label_imp



"""Functions to work with the RoiManager."""

from ij.plugin import RoiEnlarger, RoiScaler  # pylint: disable-msg=import-error
from ij.plugin.frame import RoiManager  # pylint: disable-msg=import-error


def get_roimanager():
    """Instantiate or get the IJ-RoiManager instance.

    Use to either get the current instance of the IJ RoiManager or instantiate
    it if it does not yet exist.

    Returns
    -------
    ij.plugin.frame.RoiManager
        A reference of the IJ-RoiManager.
    """
    rm = RoiManager.getInstance()
    if not rm:
        rm = RoiManager()
    return rm


def clear_ij_roi_manager(rm):
    """Delete all ROIs from the RoiManager.

    Parameters
    ----------
    rm : ij.plugin.frame.RoiManager
        A reference of the IJ-RoiManager.
    """
    rm.runCommand("reset")


def count_all_rois(rm):
    """Count the number of ROIS in the RoiManager.

    Parameters
    ----------
    rm : ij.plugin.frame.RoiManager
        A reference of the IJ-RoiManager.

    Returns
    -------
    int
        The number of ROIs in the RoiManager.
    """
    number_of_rois = rm.getCount()

    return number_of_rois


def save_rois_to_zip(rm, target, selected_rois=None):
    """Save selected ROIs in the RoiManager as zip to the target path.

    Parameters
    ----------
    rm : ij.plugin.frame.RoiManager
        A reference of the IJ-RoiManager.
    target : string
        The path to store the ROIs, e.g. /my-images/resulting_rois_subset.zip
    selected_rois : list
        selected ROIs in the RoiManager to save
    """
    if selected_rois is not None:
        rm.runCommand("Deselect")
        rm.setSelectedIndexes(selected_rois)
        rm.runCommand("save selected", target)
        rm.runCommand("Deselect")
    else:
        rm.runCommand("Save", target)


def show_all_rois_on_image(rm, imp):
    """Show all ROIs in the ROiManager on the given ImagePlus.

    Parameters
    ----------
    rm : ij.plugin.frame.RoiManager
        A reference of the IJ-RoiManager.
    imp : ij.ImagePlus
        The imp on which to show the ROIs.
    """
    rm.runCommand(imp, "Show All")


def rename_rois(rm, string):
    """Rename all ROIs to include the given string as a prefix.

    Parameters
    ----------
    rm : ij.plugin.frame.RoiManager
        A reference of the IJ-RoiManager.
    string : str
        The string to prefix the ROIs with.
    """
    number_of_rois = rm.getCount()
    for roi in range(number_of_rois):
        rm.rename(roi, string + str(roi + 1))

    rm.runCommand("UseNames", "true")


def rename_rois_by_number(rm):
    """Rename all ROIs in the RoiManager according to their index number.

    Parameters
    ----------
    rm : ij.plugin.frame.RoiManager
        A reference of the IJ-RoiManager.
    """
    number_of_rois = rm.getCount()
    for roi in range(number_of_rois):
        rm.rename(roi, str(roi + 1))


def change_roi_color(rm, color, selected_rois=None):
    """Change the color of selected / all ROIs in the RoiManager.

    Parameters
    ----------
    rm : ij.plugin.frame.RoiManager
        A reference of the IJ-RoiManager.
    color : string
        The desired color. e.g. "green", "red", "yellow", "magenta" ...
    selected_rois : list, optional
        ROIs in the RoiManager that should be changed. By default None which
        will result in all ROIs to be changed.
    """
    if selected_rois is not None:
        rm.runCommand("Deselect")
        rm.setSelectedIndexes(selected_rois)
        rm.runCommand("Set Color", color)
        rm.runCommand("Deselect")
    else:
        number_of_rois = rm.getCount()
        for roi in range(number_of_rois):
            rm.select(roi)
            rm.runCommand("Set Color", color)


def measure_in_all_rois(imp, channel, rm):
    """Perform all configured measurements in one channel of the given image.

    The choice of measured parameters is done through ImageJ's "Set
    Measurements" command.

    Parameters
    ----------
    imp : ij.ImagePlus
        The imp to measure on.
    channel : integer
        The channel to measure in (starting at 1).
    rm : ij.plugin.frame.RoiManager
        A reference of the IJ-RoiManager.
    """
    imp.setC(channel)
    rm.runCommand(imp, "Deselect")
    rm.runCommand(imp, "Measure")


def load_rois_from_zip(rm, path):
    """Load ROIs from the given zip file and add them to the RoiManager.

    Parameters
    ----------
    rm : ij.plugin.frame.RoiManager
        A reference of the IJ-RoiManager.
    path : string
        Path to the ROI zip file.
    """
    rm.runCommand("Open", path)


def enlarge_all_rois(amount_in_um, rm, pixel_size_in_um):
    """Enlarge all ROIs in the RoiManager by x scaled units.

    Parameters
    ----------
    amount_in_um : float
        The value by which to enlarge in scaled units, e.g 3.5.
    rm : ij.plugin.frame.RoiManager
        A reference of the IJ-RoiManager.
    pixel_size_in_um : float
        The pixel size, e.g. 0.65 px/um.
    """
    amount_px = amount_in_um / pixel_size_in_um
    all_rois = rm.getRoisAsArray()
    rm.reset()
    for roi in all_rois:
        enlarged_roi = RoiEnlarger.enlarge(roi, amount_px)
        rm.addRoi(enlarged_roi)


def scale_all_rois(rm, scaling_factor):
    """Inflate or shrink all ROIs in the RoiManager.

    Parameters
    ----------
    rm : ij.plugin.frame.RoiManager
        A reference of the IJ-RoiManager.
    scaling_factor : float
        The scaling factor by which to inflate (if > 1) or shrink (if < 1 ).
    """
    all_rois = rm.getRoisAsArray()
    rm.reset()
    for roi in all_rois:
        scaled_roi = RoiScaler.scale(roi, scaling_factor, scaling_factor, True)
        rm.addRoi(scaled_roi)


def select_rois_above_min_intensity(imp, channel, rm, min_intensity):
    """Select ROIs based on their intensity in a given channel of the image.

    See https://imagej.nih.gov/ij/developer/api/ij/process/ImageStatistics.html

    Parameters
    ----------
    imp : ij.ImagePlus
        The imp on which to measure.
    channel : integer
        The channel to measure in (starting at 1).
    rm : ij.plugin.frame.RoiManager
        A reference of the IJ-RoiManager.
    min_intensity : integer
        The selection criterion (lower intensity threshold).

    Returns
    -------
    list(int)
        A list of ROI index numbers fulfilling the selection criterion
        (intensity is above the threshold).
    """
    imp.setC(channel)
    all_rois = rm.getRoisAsArray()
    selected_rois = []
    for i, roi in enumerate(all_rois):
        imp.setRoi(roi)
        stats = imp.getStatistics()
        if stats.max > min_intensity:
            selected_rois.append(i)

    return selected_rois


def extract_color_of_all_rois(rm):
    """Get the color names of the ROIs in the RoiManager.

    Iterates over all ROIs and gets either their "Stroke Color" (if present) or
    their "Color" property.

    Parameters
    ----------
    rm : ij.plugin.frame.RoiManager
        A reference of the IJ-RoiManager.

    Returns
    -------
    list
        A list containing the corresponding color name strings for each ROI.
    """
    rgb_color_lookup = {
        -65536: "red",
        -65281: "magenta",
        -16711936: "green",
        -256: "yellow",
        -1: "white",
        -16776961: "blue",
        -16777216: "black",
        -14336: "orange",
        -16711681: "cyan",
    }

    all_rois = rm.getRoisAsArray()
    roi_colors = []
    for roi in all_rois:
        stroke_color = roi.getStrokeColor()
        if stroke_color:
            roi_colors.append(rgb_color_lookup[stroke_color.getRGB()])
        else:
            roi_colors.append(rgb_color_lookup[roi.getColor().getRGB()])

    return roi_colors


def add_rois_to_roimanager(
    roi_array, roi_manager, keep_rois_name, prefix, bbox=None, z_slice=None, group=None
):
    """Add all ROIs from a list to the RoiManager.

    Parameters
    ----------
    roi_array : list(ij.gui.Roi)
        List of ROIs to put in RM.
    roi_manager : ij.plugin.frame.RoiManager
        ROIManager in which to put the ROIs.
    keep_rois_name : bool
        If true, will keep the name of the ROI. Otherwise the ROI will be
        renamed using its index number.
    prefix : str
        String to prefix the name of the ROI with.
    bbox : java.awt.Rectangle, optional
        Use this bounding box to shift the ROI list, by default None.
    z_slice : int, optional
        Shift the ROI also in Z, by default None (=no shifting).
    group : int, optional
        Put the ROI into the given ROI group, by default None.
    """
    # roi_manager.reset()
    for index, roi in enumerate(roi_array):
        if not keep_rois_name:
            roi.setName(prefix + "-" + str(index))
        else:
            roi.setName(prefix + "-" + roi.getName())
        if bbox is not None:
            roi = shift_roi_by_bounding_box(roi, bbox, z_slice)
        if group is not None:
            roi.setGroup(group)
        roi_manager.addRoi(roi)


def shift_roi_by_bounding_box(roi, bbox, z_slice=None):
    """Move a ROI based on a bounding box.

    Translate one ROI based on another ROI's bounding box.

    Parameters
    ----------
    roi : ij.gui.Roi
        The ROI to be moved.
    bbox : java.awt.Rectangle
        The bounding box by which the ROI should be shifted, e.g. retrieved by
        calling `OtherRoi.getBounds()` on a ROI object.
    z_slice : int, optional
        Shift the ROI also in Z, by default None (=no shifting).
    """
    # roi_manager.reset()
    roi.setLocation(bbox.x + roi.getBounds().x, bbox.y + roi.getBounds().y)
    if z_slice is not None:
        roi.setPosition(roi.getPosition() + z_slice)
    return roi



"""Functions to work with 3D objects.

Mostly (although not exclusively) related to the [`mcib3d`][mcib3d] package.

[mcib3d]: https://mcib3d.frama.io/3d-suite-imagej/
"""

from de.mpicbg.scf.imgtools.image.create.image import ImageCreationUtilities
from de.mpicbg.scf.imgtools.image.create.labelmap import WatershedLabeling
from ij import IJ
from mcib3d.geom import Objects3DPopulation
from mcib3d.image3d import ImageHandler, ImageLabeller
from mcib3d.image3d.processing import MaximaFinder
from net.imglib2.img import ImagePlusAdapter


def population3d_to_imgplus(imp, population):
    """Make an ImagePlus from an Objects3DPopulation (2D/3D).

    Works on: 2D and 3D.

    Parameters
    ----------
    imp : ij.ImagePlus
        Original ImagePlus to derive the size of the resulting ImagePlus.
    population : mcib3d.geom.Objects3DPopulation
        Population of 3D objects used to generate the new ImagePlus.

    Returns
    -------
    ij.ImagePlus
        A newly created ImagePlus representing the labeled population.
    """
    dim = imp.getDimensions()

    # Create a new 16-bit image with the same size as the original image
    new_imp = IJ.createImage(
        "Filtered labeled stack",
        "16-bit black",
        dim[0],
        dim[1],
        1,
        dim[3],
        dim[4],
    )
    new_imp.setCalibration(imp.getCalibration())

    # Wrap the new image in an ImageHandler and draw the population
    new_img = ImageHandler.wrap(new_imp)
    population.drawPopulation(new_img)

    return new_img.getImagePlus()


def imgplus_to_population3d(imp):
    """Get an Objects3DPopulation from an ImagePlus (2D/3D).

    Works on: 2D and 3D.

    Parameters
    ----------
    imp : ij.ImagePlus
        Labeled 2D image or 3D stack used to get the population.

    Returns
    -------
    mcib3d.geom.Objects3DPopulation
        The extracted population from the image.
    """
    img = ImageHandler.wrap(imp)
    return Objects3DPopulation(img)


def segment_3d_image(imp, title=None, min_thresh=1, min_vol=None, max_vol=None):
    """Segment a 3D binary image to get a labelled stack.

    Parameters
    ----------
    imp : ij.ImagePlus
        A binary 3D stack for segmentation.
    title : str, optional
        Title of the new image. Defaults to None.
    min_thresh : int, optional
        Threshold to do segmentation, also allows for label filtering. Since the
        segmentation is happening on a binary stack, values are either 0 or 255,
        so using 0 allows to discard only the background.  Defaults to 1.
    min_vol : int, optional
        Minimum volume (in voxels) under which objects get filtered.
        Defaults to None.
    max_vol : int, optional
        Maximum volume (in voxels) above which objects get filtered.
        Defaults to None.

    Returns
    -------
    ij.ImagePlus
        A labelled 3D ImagePlus.
    """
    cal = imp.getCalibration()

    # Wrap through ImageHandler and apply thresholding
    img = ImageHandler.wrap(imp)
    img = img.threshold(min_thresh, False, False)

    labeler = ImageLabeller()
    if min_vol:
        labeler.setMinSizeCalibrated(min_vol, img)
    if max_vol:
        labeler.setMaxSizeCalibrated(max_vol, img)

    # Generate labelled segmentation
    seg = labeler.getLabels(img)
    seg.setScale(cal.pixelWidth, cal.pixelDepth, cal.getUnits())
    if title:
        seg.setTitle(title)

    return seg.getImagePlus()


def maxima_finder_3d(imp, min_threshold=0, noise=100, rxy=1.5, rz=1.5):
    """Find local maxima in a 3D image.

    This function identifies local maxima in a 3D image using a specified minimum threshold and noise level.
    The radii for the maxima detection can be set independently for the x/y and z dimensions.

    Parameters
    ----------
    imp : ij.ImagePlus
        The input 3D image in which to find local maxima.
    min_threshold : int, optional
        The minimum intensity threshold for maxima detection. Default is 0.
    noise : int, optional
        The noise tolerance level for maxima detection. Default is 100.
    rxy : float, optional
        The radius for maxima detection in the x and y dimensions. Default is 1.5.
    rz : float, optional
        The radius for maxima detection in the z dimension. Default is 1.5.

    Returns
    -------
    ij.ImagePlus
        An ImagePlus object containing the detected maxima as peaks.
    """
    # Wrap the input ImagePlus into an ImageHandler
    img = ImageHandler.wrap(imp)

    # Duplicate the image and apply a threshold cut-off
    thresholded = img.duplicate()
    thresholded.thresholdCut(min_threshold, False, True)

    # Initialize the MaximaFinder with the thresholded image and noise level
    maxima_finder = MaximaFinder(thresholded, noise)

    # Set the radii for maxima detection in x/y and z dimensions
    maxima_finder.setRadii(rxy, rz)

    # Retrieve the image peaks as an ImageHandler
    img_peaks = maxima_finder.getImagePeaks()

    # Convert the ImageHandler peaks to an ImagePlus
    imp_peaks = img_peaks.getImagePlus()

    # Set the calibration of the peaks image to match the input image
    imp_peaks.setCalibration(imp.getCalibration())

    # Set the title of the peaks image
    imp_peaks.setTitle("Peaks")

    return imp_peaks


def seeded_watershed(imp_binary, imp_peaks, threshold=10):
    """Perform a seeded watershed segmentation on a binary image using seed points.

    This function applies a watershed segmentation to a binary image using seed points provided in another image.
    An optional threshold can be specified to control the segmentation process.

    Parameters
    ----------
    imp_binary : ij.ImagePlus
        The binary image to segment.
    imp_peaks : ij.ImagePlus
        The image containing the seed points for the watershed segmentation.
    threshold : float, optional
        The threshold value to use for the segmentation. Default is 10.

    Returns
    -------
    ij.ImagePlus
        The segmented image with labels.
    """

    img = ImagePlusAdapter.convertFloat(imp_binary)
    img_seed = ImagePlusAdapter.convertFloat(imp_peaks).copy()

    if threshold:
        watersheded_result = WatershedLabeling.watershed(img, img_seed, threshold)
    else:
        watersheded_result = WatershedLabeling.watershed(img, img_seed)

    return ImageCreationUtilities.convertImgToImagePlus(
        watersheded_result,
        "Label image",
        "",
        imp_binary.getDimensions(),
        imp_binary.getCalibration(),
    )



"""Functions allowing to interact with an OMERO server.

Contains helpers to parse URLs and / or OMERO image IDs, connect to OMERO and
fetch images from the server.

Requires both the [`simple-omero-client`][simple-omero-client] and the
[`omero-insight`][omero-insight] JARs to be installed.

Most of the functions will use the [`simple-omero-client`][simple-omero-client]
to interact with the OMERO server. However, there are still some that
requires the [`omero-insight`][omero-insight] plugin to read metadata.

[simple-omero-client]: https://github.com/GReD-Clermont/simple-omero-client
[omero-insight]: https://github.com/ome/omero-insight
"""

from fr.igred.omero import Client
from fr.igred.omero.annotations import (
    MapAnnotationWrapper,
    TableWrapper,
)
from fr.igred.omero.roi import ROIWrapper
from java.lang import Long
from java.text import SimpleDateFormat
from java.util import ArrayList
from omero.cmd import OriginalMetadataRequest
from omero.gateway.model import TableData, TableDataColumn


def parse_url(client, omero_str):
    """Parse an OMERO URL / image IDs into the respective ImageWrapper objects.

    Assemble a list of ImageWrapper objects from one of the following inputs:

    - An OMERO URL (as retrieved e.g. from OMERO.web).
    - One or more OMERO image IDs.
    - An OMERO Dataset ID.

    Parameters
    ----------
    client : fr.igred.omero.Client
        Client used for login to OMERO.
    omero_str : str
        Either an URL from OMERO or image IDs separated by commas.

    Returns
    -------
    list(fr.igred.omero.repository.ImageWrapper)
        List of ImageWrappers parsed from the string.

    Examples
    --------
    >>> from fr.igred.omero import Client
    >>> client = Client()
    >>> OMERO_LINK = "123456"
    >>> img_wrappers = omerotools.parse_url(client, OMERO_LINK)
    >>> for wrapper in img_wrappers:
    >>>    imp = wpr.toImagePlus(client)
    """
    image_ids = []
    dataset_ids = []
    image_wpr_list = []

    # Check if the url is a dataset
    if "dataset-" in omero_str:
        # If there are multiple datasets
        if "|" in omero_str:
            parts = omero_str.split("|")
            for part in parts:
                if "dataset-" in part:
                    image_ids.extend(
                        [
                            image
                            for image in client.getDataset(
                                Long(part.split("dataset-")[1].split("/")[0])
                            ).getImages()
                        ]
                    )
                    dataset_id = Long(part.split("dataset-")[1].split("/")[0])
                    dataset_ids.append(dataset_id)
        else:
            image_ids.extend(
                [
                    image
                    for image in client.getDataset(
                        Long(omero_str.split("dataset-")[1].split("/")[0])
                    ).getImages()
                ]
            )
            # If there is only one dataset
            dataset_id = Long(omero_str.split("dataset-")[1].split("/")[0])
            dataset_ids.append(dataset_id)

        # Get the images from the dataset
        for dataset_id in dataset_ids:
            dataset_wpr = client.getDataset(dataset_id)
            image_wpr_list.extend(dataset_wpr.getImages())

        return image_wpr_list

    # Check if the url is an image
    elif "image-" in omero_str:
        image_ids = omero_str.split("image-")
        image_ids.pop(0)
        image_ids = [s.split("%")[0].replace("|", "") for s in image_ids]
    else:
        image_ids = (
            [s.split("%")[0].replace("|", "") for s in omero_str.split("image-")[1:]]
            if "image-" in omero_str
            else omero_str.split(",")
        )
        # If it is a list of IDs separated by commas
        image_ids = omero_str.split(",")

    return [client.getImage(Long(image_id)) for image_id in image_ids]


def connect(host, port, username, password):
    """Connect to OMERO using the credentials provided.

    Parameters
    ----------
    host : str
        The address (FQDN or IP) of the OMERO server.
    port : int
        The port number for the OMERO server.
    username : str
        The username for authentication.
    password : str
        The password for authentication.

    Returns
    -------
    fr.igred.omero.Client
        A Client object representing the connection to the OMERO server.
    """
    # Create a new OMERO client
    client = Client()

    # Connect to the OMERO server using provided credentials
    client.connect(host, port, username, password)

    # Return the connected client
    return client


def fetch_image(client, image_id):
    """Fetch an image from an OMERO server and open it as an ImagePlus.

    Parameters
    ----------
    client : fr.igred.omero.Client
        The client object used to connect to the OMERO server.
    image_id : int
        The ID of the image to fetch.

    Returns
    -------
    ij.ImagePlus
        The fetched image as an ImagePlus.
    """

    # Fetch the image from the OMERO server
    image_wpr = client.getImage(Long(image_id))

    # Convert the image to an ImagePlus
    return image_wpr.toImagePlus()


def upload_image_to_omero(user_client, path, dataset_id):
    """Upload an image to OMERO.

    Parameters
    ----------
    user_client : fr.igred.omero.Client
        The client object used to connect to the OMERO server.
    path : str
        Path of the file to upload back to OMERO.
    dataset_id : Long
        ID of the dataset where to upload the file.

    Returns
    -------
    Long
        ID of the uploaded image
    """
    return user_client.getDataset(Long(dataset_id)).importImage(user_client, path)[0]


def add_keyvalue_annotation(client, repository_wpr, annotations, header):
    """Add an annotation to an OMERO object.

    Parameters
    ----------
    client : fr.igred.omero.Client
        The client object used to connect to the OMERO server.
    repository_wpr : fr.igred.omero.repositor.GenericRepositoryObjectWrapper
        Wrapper to the object for the anotation.
    annotations : dict
        Dictionary with the annotation to add.
    header : str
        Name for the annotation header.
    """
    # for pair in dict:
    #     result.add
    map_annotation_wpr = MapAnnotationWrapper(annotations)
    map_annotation_wpr.setNameSpace(header)
    repository_wpr.addMapAnnotation(client, map_annotation_wpr)


def delete_keyvalue_annotations(user_client, object_wrapper):
    """Delete annotations linked to object.

    Parameters
    ----------
    user_client : fr.igred.omero.Client
        Client used for login to OMERO.
    object_wrapper : fr.igred.omero.repositor.GenericRepositoryObjectWrapper
        Wrapper to the object for the anotation.

    """
    kv_pairs = object_wrapper.getMapAnnotations(user_client)
    user_client.delete(kv_pairs)


def find_dataset(client, dataset_id):
    """Retrieve a dataset (wrapper) from the OMERO server.

    Parameters
    ----------
    client : fr.igred.omero.Client
        The client object used to connect to the OMERO server.
    dataset_id : int
        The ID of the dataset to retrieve.

    Returns
    -------
    fr.igred.omero.repositor.DatasetWrapper
        The dataset wrapper retrieved from the server.
    """
    # Fetch the dataset from the OMERO server using the provided dataset ID
    return client.getDataset(Long(dataset_id))


def get_acquisition_metadata(user_client, image_wpr):
    """Get acquisition metadata from OMERO based on an image wrapper.

    Parameters
    ----------
    user_client : fr.igred.omero.Client
        Client used for login to OMERO
    image_wpr : fr.igred.omero.repositor.ImageWrapper
        Wrapper to the image for the metadata

    Returns
    -------
    dict

        {
            objective_magnification : float,
            objective_na : float,
            acquisition_date : str,
            acquisition_date_number : str,
        }
    """
    ctx = user_client.getCtx()
    instrument_data = (
        user_client.getGateway()
        .getMetadataService(ctx)
        .loadInstrument(image_wpr.asDataObject().getInstrumentId())
    )
    objective_data = instrument_data.copyObjective().get(0)
    metadata = {}

    metadata["objective_magnification"] = (
        objective_data.getNominalMagnification().getValue()
        if objective_data.getNominalMagnification() is not None
        else 0
    )
    metadata["objective_na"] = (
        objective_data.getLensNA().getValue()
        if objective_data.getLensNA() is not None
        else 0
    )

    if image_wpr.getAcquisitionDate() is None:
        if image_wpr.asDataObject().getFormat() == "ZeissCZI":
            field = "Information|Document|CreationDate"
            date_field = get_info_from_original_metadata(user_client, image_wpr, field)
            metadata["acquisition_date"] = date_field.split("T")[0]
            metadata["acquisition_date_number"] = int(
                metadata["acquisition_date"].replace("-", "")
            )
        else:
            metadata["acquisition_date"] = "NA"
            metadata["acquisition_date_number"] = 0
    else:
        sdf = SimpleDateFormat("yyyy-MM-dd")
        metadata["acquisition_date"] = sdf.format(image_wpr.getAcquisitionDate())
        metadata["acquisition_date_number"] = int(
            metadata["acquisition_date"].replace("-", "")
        )

    return metadata


def get_info_from_original_metadata(user_client, image_wpr, field):
    """Retrieve information from the original metadata (as opposed to OME-MD).

    In some cases not all information is parsed correctly by BF and has to be
    recovered / identified directly from the *original* metadata. This function
    extracts the corresponding value based on the field identifier.

    Parameters
    ----------
    user_client : fr.igred.omero.Client
        Client used for login to OMERO
    image_wpr : fr.igred.omero.repositor.ImageWrapper
        Wrapper to the image
    field : str
        Field to look for in the original metadata. Needs to be found beforehand.

    Returns
    -------
    str
        Value of the field
    """
    omr = OriginalMetadataRequest(Long(image_wpr.getId()))
    cmd = user_client.getGateway().submit(user_client.getCtx(), omr)
    rsp = cmd.loop(5, 500)
    gm = rsp.globalMetadata
    return gm.get(field).getValue()


def create_table_columns(headings):
    """Create OMERO table headings from a list of column names.

    Parameters
    ----------
    headings : list(str)
        List of columns names.

    Returns
    -------
    list(omero.gateway.model.TableDataColumn)
        List of columns formatted to be uploaded to OMERO.
    """
    table_columns = []
    # populate the headings
    for h in range(len(headings)):
        heading = headings.keys()[h]
        type = headings.values()[h]
        # OMERO.tables queries don't handle whitespace well
        heading = heading.replace(" ", "_")
        # title_heading = ["Slice", "Label"]
        table_columns.append(TableDataColumn(heading, h, type))
    # table_columns.append(TableDataColumn("Image", size, ImageData))
    return table_columns


def upload_array_as_omero_table(user_client, table_title, data, columns, image_wpr):
    """Upload a table to OMERO from a list of lists.

    Parameters
    ----------
    user_client : fr.igred.omero.Client
        Client used for login to OMERO
    table_title : str
        Title of the table to be uploaded.
    data : list(list())
        List of lists of results to upload
    columns : list(str)
        List of columns names
    image_wpr : fr.igred.omero.repositor.ImageWrapper
        Wrapper to the image to be uploaded

    Examples
    --------
    >>> from fr.igred.omero import Client
    >>> from java.lang import String, Double, Long
    ...
    >>> client = Client()  # connect to OMERO
    >>> client.connect("omero.example.org", 4064, "username", "password")
    ...
    >>> image_wpr = client.getImage(Long(123456))  # get an image
    ...
    >>> columns = {  # prepare column definitions (name-type pairs)
    ...     "Row_ID": Long,
    ...     "Cell_Area": Double,
    ...     "Cell_Type": String,
    ... }
    ...
    >>> data = [  # prepare data (list of rows, each row is a list of values)
    ...     [1, 250.5, "Neuron"],
    ...     [2, 180.2, "Astrocyte"],
    ...     [3, 310.7, "Neuron"],
    ... ]
    ...
    >>> upload_array_as_omero_table(
    ...     client, "Cell Measurements", data, columns, image_wpr
    ... )
    """

    dataset_wpr = image_wpr.getDatasets(user_client)[0]

    table_columns = create_table_columns(columns)
    table_data = TableData(table_columns, data)
    table_wpr = TableWrapper(table_data)
    table_wpr.setName(table_title)
    dataset_wpr.addTable(user_client, table_wpr)


def save_rois_to_omero(user_client, image_wpr, rm):
    """Save ROIs to OMERO linked to the image.

    Parameters
    ----------
    user_client : fr.igred.omero.Client
        Client used for login to OMERO
    image_wpr : fr.igred.omero.repositor.ImageWrapper
        Wrapper to the image for the ROIs
    rm : ij.plugin.frame.RoiManager
        ROI Manager containing the ROIs

    """
    rois_list = rm.getRoisAsArray()
    rois_arraylist = ArrayList(len(rois_list))
    for roi in rois_list:
        rois_arraylist.add(roi)
    rois_to_upload = ROIWrapper.fromImageJ(rois_arraylist)
    image_wpr.saveROIs(user_client, rois_to_upload)



"""Bio-Formats related helper functions.

NOTE: this is *NOT* about using [python-bioformats][1] but rather about calling
the corresponding functions provided by ImageJ.

[1]: https://pypi.org/project/python-bioformats/

"""

# Mosts imports will fail with plain C-Python / pylint:
# pylint: disable-msg=import-error

import os

from ij import IJ

from ..log import LOG as log
from ..pathtools import gen_name_from_orig
from ._loci import (
    BF,
    DynamicMetadataOptions,
    ImageReader,
    ImporterOptions,
    Memoizer,
    MetadataTools,
    ZeissCZIReader,
    Region,
)


class ImageMetadata(object):
    """A class to store metadata information from an image.

    This class stores metadata information extracted from an image file, such as image dimensions,
    pixel dimensions, and calibration units. It provides a method to convert the attributes to a
    dictionary and a string representation of the object.

    Attributes
    ----------
    unit_width : float or None
        Physical width of a pixel in the given unit.
    unit_height : float or None
        Physical height of a pixel in the given unit.
    unit_depth : float or None
        Physical depth of a voxel in the given unit.
    pixel_width : int or None
        Width of the image in pixels.
    pixel_height : int or None
        Height of the image in pixels.
    slice_count : int or None
        Number of Z-slices in the image.
    channel_count : int or None
        Number of channels in the image.
    timepoints_count : int or None
        Number of timepoints in the image.
    dimension_order : str or None
        Order of dimensions (e.g., "XYZCT").
    pixel_type : str or None
        Data type of the pixel values (e.g., "uint16").

    Examples
    --------
    >>> metadata = ImageMetadata(
    ...     unit_width=0.1,
    ...     unit_height=0.1
    ...     )
    >>> print(metadata)
    <ImageMetadata(unit_width=0.1, unit_height=0.1, ...)>
    """

    def __init__(
        self,
        unit_width=None,
        unit_height=None,
        unit_depth=None,
        unit=None,
        pixel_width=None,
        pixel_height=None,
        slice_count=None,
        channel_count=None,
        timepoints_count=None,
        dimension_order=None,
        pixel_type=None,
    ):
        self.unit_width = unit_width
        self.unit_height = unit_height
        self.unit_depth = unit_depth
        self.unit = unit
        self.pixel_width = pixel_width
        self.pixel_height = pixel_height
        self.slice_count = slice_count
        self.channel_count = channel_count
        self.timepoints_count = timepoints_count
        self.dimension_order = dimension_order
        self.pixel_type = pixel_type

    def to_dict(self):
        """Convert the object attributes to a dictionary.

        Returns
        -------
        dict
            A dictionary representation of the object attributes.
        """
        return self.__dict__


class StageMetadata(object):
    """A class to store stage coordinates and calibration metadata for a set of images.

    Attributes
    ----------
    dimensions : int
        Number of dimensions (2D or 3D).
    stage_coordinates_x : list of float
        Absolute stage x-coordinates.
    stage_coordinates_y : list of float
        Absolute stage y-coordinates.
    stage_coordinates_z : list of float
        Absolute stage z-coordinates.
    relative_coordinates_x : list of float
        Relative stage x-coordinates in pixels.
    relative_coordinates_y : list of float
        Relative stage y-coordinates in pixels.
    relative_coordinates_z : list of float
        Relative stage z-coordinates in pixels.
    image_calibration : list of float
        Calibration values for x, y, and z in unit/px.
    calibration_unit : str
        Unit used for image calibration.
    image_dimensions_czt : list of int
        Number of images in dimensions (channels, z-slices, timepoints).
    series_names : list of str
        Names of all series in the image files.
    max_size : list of float
        Maximum physical size (x/y/z) across all files.
    """

    def __init__(
        self,
        dimensions=2,
        stage_coordinates_x=None,
        stage_coordinates_y=None,
        stage_coordinates_z=None,
        relative_coordinates_x=None,
        relative_coordinates_y=None,
        relative_coordinates_z=None,
        image_calibration=None,
        calibration_unit="unknown",
        image_dimensions_czt=None,
        series_names=None,
        max_size=None,
    ):
        self.dimensions = dimensions
        self.stage_coordinates_x = stage_coordinates_x or []
        self.stage_coordinates_y = stage_coordinates_y or []
        self.stage_coordinates_z = stage_coordinates_z or []
        self.relative_coordinates_x = relative_coordinates_x or []
        self.relative_coordinates_y = relative_coordinates_y or []
        self.relative_coordinates_z = relative_coordinates_z or []
        self.image_calibration = image_calibration or [1.0, 1.0, 1.0]
        self.calibration_unit = calibration_unit or "unknown"
        self.image_dimensions_czt = image_dimensions_czt or [1, 1, 1]
        self.series_names = series_names or []
        self.max_size = max_size or [1.0, 1.0, 1.0]

    def __repr__(self):
        """Return a string representation of the object."""
        return "<StageMetadata({})>".format(
            ", ".join("{}={}".format(k, v) for k, v in self.__dict__.items())
        )


def import_image(
    filename,
    color_mode="color",
    split_c=False,
    split_z=False,
    split_t=False,
    series_number=None,
    c_start=None,
    c_end=None,
    c_interval=None,
    z_start=None,
    z_end=None,
    z_interval=None,
    t_start=None,
    t_end=None,
    t_interval=None,
    region=None,
):
    """Open an image file using the Bio-Formats importer.

    Parameters
    ----------
    filename : str
        The full path to the file to be imported through Bio-Formats.
    color_mode : str, optional
        The color mode to be used for the resulting ImagePlus, one of 'color',
        'composite', 'gray' and 'default'.
    split_c : bool, optional
        Whether to split the channels into separate ImagePlus objects.
    split_z : bool, optional
        Whether to split the z-slices into separate ImagePlus objects.
    split_t : bool, optional
        Whether to split the time points into separate ImagePlus objects.
    series_number : int, optional
        open a specific Bio-Formats series
    c_start : int, optional
        only import a subset of channel starting with this one. Requires to set
        c_end and c_interval.
    c_end : int, optional
        only import channel(s) ending with this one. Requires to set c_start and
        c_interval.
    c_interval : int, optional
        only import a subset of channel with this interval. Requires to set
        c_start and c_end.
    z_start : int, optional
        only import a subset of planes starting with this one. Requires to set
        z_end and z_interval.
    z_end : int, optional
        only import a subset of planes ending with this one. Requires to set
        z_start and z_interval.
    z_interval : int, optional
        only import a subset of planes with this interval. Requires to set
        z_start and z_end.
    t_start : int, optional
        only import a subset of time points starting with this one. Requires to
        set t_end and t_interval.
    t_end : int, optional
        only import a subset of time points ending with this one. Requires to
        set t_start and t_interval.
    t_interval : int, optional
        only import a subset of time points with this interval. Requires to set
        t_start and t_end.
    region : list, optional
        Bio-Formats crop region, by default None.
        Format: `[start_x, start_y, width_in_px, height_in_px]`.

    Returns
    -------
    list(ij.ImagePlus)
        A list of ImagePlus objects resulting from the import.
    """
    options = ImporterOptions()
    mode = {
        "color": ImporterOptions.COLOR_MODE_COLORIZED,
        "composite": ImporterOptions.COLOR_MODE_COMPOSITE,
        "gray": ImporterOptions.COLOR_MODE_GRAYSCALE,
        "default": ImporterOptions.COLOR_MODE_DEFAULT,
    }
    options.setColorMode(mode[color_mode])
    options.setSplitChannels(split_c)
    options.setSplitFocalPlanes(split_z)
    options.setSplitTimepoints(split_t)
    options.setId(filename)
    if series_number is not None:
        options.setSeriesOn(series_number, True)

    if c_start is not None:
        if series_number is None:
            series_number = 0
        options.setSpecifyRanges(True)
        options.setCBegin(series_number, c_start)
        options.setCEnd(series_number, c_end)
        options.setCStep(series_number, c_interval)

    if z_start is not None:
        if series_number is None:
            series_number = 0
        options.setSpecifyRanges(True)
        options.setZBegin(series_number, z_start)
        options.setZEnd(series_number, z_end)
        options.setZStep(series_number, z_interval)

    if t_start is not None:
        if series_number is None:
            series_number = 0
        options.setSpecifyRanges(True)
        options.setTBegin(series_number, t_start)
        options.setTEnd(series_number, t_end)
        options.setTStep(series_number, t_interval)

    if region is not None:
        options.setCrop(True)
        options.setCropRegion(
            series_number, Region(region[0], region[1], region[2], region[3])
        )

    log.info("Reading [%s]", filename)
    orig_imps = BF.openImagePlus(options)
    log.debug("Opened [%s] %s", filename, type(orig_imps))
    return orig_imps


def export(imp, filename, overwrite=False):
    """Export an ImagePlus object to a given file.

    Parameters
    ----------
    imp : ij.ImagePlus
        The ImagePlus object to be exported by Bio-Formats.
    filename : str
        The output filename, may include a full path.
    overwrite : bool
        A switch to indicate existing files should be overwritten. Default is to
        keep existing files, in this case an IOError is raised.
    """
    log.info("Exporting to [%s]", filename)
    suffix = filename[-3:].lower()
    try:
        unit = imp.calibration.unit
        log.debug("Detected calibration unit: %s", unit)
    except Exception as err:
        log.error("Unable to detect spatial unit: %s", err)
        raise RuntimeError("Error detecting image calibration: %s" % err)
    if unit == "pixel" and (suffix == "ics" or suffix == "ids"):
        log.warn(
            "Forcing unit to be 'm' instead of 'pixel' to avoid "
            "Bio-Formats 6.0.x Exporter bug!"
        )
        imp.calibration.unit = "m"
    if os.path.exists(filename):
        if not overwrite:
            raise IOError("file [%s] already exists!" % filename)
        log.debug("Removing existing file [%s]...", filename)
        os.remove(filename)

    IJ.run(imp, "Bio-Formats Exporter", "save=[" + filename + "]")
    log.debug("Exporting finished.")


def export_using_orig_name(imp, path, orig_name, tag, suffix, overwrite=False):
    """Export an image to a given path, deriving the name from the input file.

    The input filename is stripped to its pure file name, without any path or
    suffix components, then an optional tag (e.g. "-avg") and the new format
    suffix is added.

    Parameters
    ----------
    imp : ij.ImagePlus
        The ImagePlus object to be exported by Bio-Formats.
    path : str or object that can be cast to a str
        The output path.
    orig_name : str or object that can be cast to a str
        The input file name, may contain arbitrary path components.
    tag : str
        An optional tag to be added at the end of the new file name, can be used
        to denote information like "-avg" for an average projection image.
    suffix : str
        The new file name suffix, which also sets the file format for BF.
    overwrite : bool
        A switch to indicate existing files should be overwritten.

    Returns
    -------
    str
        The full name of the exported file.
    """
    out_file = gen_name_from_orig(path, orig_name, tag, suffix)
    export(imp, out_file, overwrite)
    return out_file


def get_reader(path_to_file, setFlattenedResolutions=False):
    """Get a Bio-Formats ImageReader for the specified file.

    Parameters
    ----------
    path_to_file : str
        The full path to the image file.
    setFlattenedResolutions : bool, optional
        Whether to flatten resolutions in the ImageReader (default: False).

    Returns
    -------
    ImageReader
        A configured ImageReader instance for the specified file.
    """
    reader = ImageReader()
    ome_meta = MetadataTools.createOMEXMLMetadata()
    reader.setMetadataStore(ome_meta)
    m = DynamicMetadataOptions()
    m.setBoolean(ZeissCZIReader.ALLOW_AUTOSTITCHING_KEY, False)
    reader.setMetadataOptions(m)
    reader.setId(str(path_to_file))
    return reader, ome_meta


def get_series_info_from_ome_metadata(path_to_file, skip_labels=False):
    """Get the Bio-Formats series information from a file on disk.

    Useful to access specific images in container formats like .czi, .nd2, .lif...

    Parameters
    ----------
    path_to_file : str
        The full path to the image file.
    skip_labels : bool, optional
        If True, excludes label and macro images from the series count (default: False).

    Returns
    -------
    tuple
        A tuple containing:
        - int: The number of Bio-Formats series detected (excluding labels if skip_labels=True)
        - list or range: Series indices. If skip_labels=True, returns filtered list of indices,
          otherwise returns range(series_count)

    Examples
    --------
    >>> count, indices = get_series_info_from_ome_metadata("image.czi")
    >>> count, indices = get_series_info_from_ome_metadata("image.nd2", skip_labels=True)
    """

    reader, ome_meta = get_reader(path_to_file, skip_labels)
    series_count = reader.getSeriesCount()
    if not skip_labels:
        # If we are not skipping labels, return the full range
        return series_count, range(series_count)

    series_ids = []
    series_names = []
    x = 0
    y = 0
    for i in range(series_count):
        reader.setSeries(i)

        if reader.getSizeX() > x and reader.getSizeY() > y:
            name = ome_meta.getImageName(i)

            if name not in ["label image", "macro image"]:
                series_ids.append(i)
                series_names.append(name)

        x = reader.getSizeX()
        y = reader.getSizeY()

    print(series_names)
    return len(series_ids), series_ids


def write_bf_memoryfile(path_to_file):
    """Write a BF memo-file so subsequent access to the same file is faster.

    The Bio-Formats memo-file is written next to the image file (i.e. in the
    same folder as the given file).

    Parameters
    ----------
    path_to_file : str
        The full path to the image file.
    """
    reader = Memoizer(ImageReader())
    reader.setId(path_to_file)
    reader.close()


def get_metadata_from_file(path_to_image):
    """Extract metadata from an image file using Bio-Formats.

    This function reads an image file using the Bio-Formats library and extracts
    various metadata properties including physical dimensions, pixel dimensions,
    and other image characteristics.

    Parameters
    ----------
    path_to_image : str or pathlib.Path
        Path to the image file from which metadata should be extracted.

    Returns
    -------
    ImageMetadata
        An instance of `imcflibs.imagej.bioformats.ImageMetadata` containing the extracted metadata.
    """

    reader, ome_meta = get_reader(path_to_image)

    metadata = ImageMetadata(
        unit_width=ome_meta.getPixelsPhysicalSizeX(0).value(),
        unit_height=ome_meta.getPixelsPhysicalSizeY(0).value(),
        unit_depth=ome_meta.getPixelsPhysicalSizeZ(0).value(),
        unit=ome_meta.getPixelsPhysicalSizeX(0).unit().symbol,
        pixel_width=ome_meta.getPixelsSizeX(0),
        pixel_height=ome_meta.getPixelsSizeY(0),
        slice_count=ome_meta.getPixelsSizeZ(0),
        channel_count=ome_meta.getPixelsSizeC(0),
        timepoints_count=ome_meta.getPixelsSizeT(0),
        dimension_order=ome_meta.getPixelsDimensionOrder(0),
        pixel_type=ome_meta.getPixelsType(0),
    )
    reader.close()

    return metadata


def get_stage_coords(filenames):
    """Get stage coordinates and calibration for a given list of images.

    Parameters
    ----------
    filenames : list of str
        List of image filepaths.

    Returns
    -------
    StageMetadata
        An object containing extracted stage metadata.
    """
    # Initialize lists to store stage coordinates and series names
    stage_coordinates_x = []
    stage_coordinates_y = []
    stage_coordinates_z = []
    series_names = []

    # Intiialize default values
    dimensions = 2
    image_calibration = []
    calibration_unit = "unknown"
    image_dimensions_czt = []
    max_size = []

    # Initialize max_size variables to track the maximums
    max_phys_size_x = 0.0
    max_phys_size_y = 0.0
    max_phys_size_z = 0.0

    for counter, image in enumerate(filenames):
        reader, ome_meta = get_reader(image)
        series_count = reader.getSeriesCount()

        # Process only the first image to get values not dependent on series
        if counter == 0:
            frame_size_x = reader.getSizeX()
            frame_size_y = reader.getSizeY()
            frame_size_z = reader.getSizeZ()
            frame_size_c = reader.getSizeC()
            frame_size_t = reader.getSizeT()

            dimensions = 2 if frame_size_z == 1 else 3

            # Retrieve physical size coordinates safely
            phys_size_x = getattr(
                ome_meta.getPixelsPhysicalSizeX(0), "value", lambda: 1.0
            )()
            phys_size_y = getattr(
                ome_meta.getPixelsPhysicalSizeY(0), "value", lambda: 1.0
            )()
            phys_size_z = getattr(
                ome_meta.getPixelsPhysicalSizeZ(0), "value", lambda: None
            )()

            z_interval = phys_size_z if phys_size_z is not None else 1.0

            # Handle missing Z calibration
            if phys_size_z is None and frame_size_z > 1:
                first_plane = getattr(
                    ome_meta.getPlanePositionZ(0, 0), "value", lambda: 0
                )()
                next_plane_index = frame_size_c + frame_size_t - 1
                second_plane = getattr(
                    ome_meta.getPlanePositionZ(0, next_plane_index), "value", lambda: 0
                )()
                z_interval = abs(first_plane - second_plane)

            image_calibration = [phys_size_x, phys_size_y, z_interval]
            calibration_unit = (
                getattr(
                    ome_meta.getPixelsPhysicalSizeX(0).unit(),
                    "getSymbol",
                    lambda: "unknown",
                )()
                if phys_size_x
                else "unknown"
            )
            image_dimensions_czt = [frame_size_c, frame_size_z, frame_size_t]

        reader.close()

        for series in range(series_count):
            if ome_meta.getImageName(series) == "macro image":
                continue

            if series_count > 1 and not str(image).endswith(".vsi"):
                series_names.append(ome_meta.getImageName(series))
            else:
                series_names.append(os.path.basename(str(image)))
            current_position_x = getattr(
                ome_meta.getPlanePositionX(series, 0), "value", lambda: 0
            )()
            current_position_y = getattr(
                ome_meta.getPlanePositionY(series, 0), "value", lambda: 0
            )()
            current_position_z = getattr(
                ome_meta.getPlanePositionZ(series, 0), "value", lambda: 1.0
            )()

            max_phys_size_x = max(
                max_phys_size_x, ome_meta.getPixelsPhysicalSizeX(series).value()
            )
            max_phys_size_y = max(
                max_phys_size_y, ome_meta.getPixelsPhysicalSizeY(series).value()
            )
            max_phys_size_z = max(
                max_phys_size_z,
                ome_meta.getPixelsPhysicalSizeZ(series).value()
                if phys_size_z
                else z_interval,
            )

            stage_coordinates_x.append(current_position_x)
            stage_coordinates_y.append(current_position_y)
            stage_coordinates_z.append(current_position_z)

    max_size = [max_phys_size_x, max_phys_size_y, max_phys_size_z]

    relative_coordinates_x_px = [
        (stage_coordinates_x[i] - stage_coordinates_x[0]) / (phys_size_x or 1.0)
        for i in range(len(stage_coordinates_x))
    ]
    relative_coordinates_y_px = [
        (stage_coordinates_y[i] - stage_coordinates_y[0]) / (phys_size_y or 1.0)
        for i in range(len(stage_coordinates_y))
    ]
    relative_coordinates_z_px = [
        (stage_coordinates_z[i] - stage_coordinates_z[0]) / (z_interval or 1.0)
        for i in range(len(stage_coordinates_z))
    ]

    return StageMetadata(
        dimensions=dimensions,
        stage_coordinates_x=stage_coordinates_x,
        stage_coordinates_y=stage_coordinates_y,
        stage_coordinates_z=stage_coordinates_z,
        relative_coordinates_x=relative_coordinates_x_px,
        relative_coordinates_y=relative_coordinates_y_px,
        relative_coordinates_z=relative_coordinates_z_px,
        image_calibration=image_calibration,
        calibration_unit=calibration_unit,
        image_dimensions_czt=image_dimensions_czt,
        series_names=series_names,
        max_size=max_size,
    )
